# FILE: buildlib.tcl
# A library of Tcl procs which supports the script ../build.tcl .
#
######################################################################
# CONFIGURATION SUPPORT
######################################################################
# proc Default var val
#  If var already exists, do nothing.  If not, set the value of var to val.
#  Returns whatever value var ends up with.
proc Default {varref val} {
  upvar $varref var
  if {![info exists var]} {
    set var $val
  }
  return $var
}

# proc EvaluateScript script
#	A more elaborate "source" which includes calling information in
#	its error messages.
proc EvaluateScript {script} {
  if [catch {uplevel #0 source $script} msg] {
    return -code error "[info script]: Unable to evaluate $script .\n$msg"
  }
}

# proc AllDefined list
#	Returns true iff all variables in the list are set.
proc AllDefined {list} {
    set result 1
    foreach varref $list {
        upvar $varref var
        set result [expr $result && [info exists var]]
    }
    return $result
}

# proc CommaList list
#	Returns a string containing the elements of list separated by
#	commas and a final "and".
proc CommaList {list} {
    set last [expr [llength $list] - 1]
    set final [lindex $list $last]
    lreplace $list $last $last "and $final"
    return [join $list ", "]
}

# proc ReportMissingValues varlist
#	Print a report to stderr that varlist includes an undefined
#	variable, and die.
proc ReportMissingValues {varlist} {
    set caller [info script]
    puts stderr "$caller: At least one of the variables"
    puts stderr "$caller: [CommaList $varlist]"
    puts stderr "$caller: could not be supplied with a default value."
    puts stderr "$caller: Please see ./build/config.tcl for"
    puts stderr "$caller: configuration instructions."
    flush stderr
    exit 1
}

proc AppendUnique {listref value} {
  upvar $listref list
  if {[lsearch -exact $list $value] < 0} {
    lappend list $value
  }
}

proc AppendUniqueExistingReadableDirectory {listref dirref} {
  upvar $listref list
  upvar $dirref dir
  if [info exists dir] {
    if {[file isdirectory $dir] && [file readable $dir]} {
      AppendUnique list $dir
    }
  }
}

# Proc to extract a recursive list of all #include ".*" files
# Results are stored in global var ArrInclDep.
# NOTE 1: Does not search into '#include <.*>' (system) files.
# NOTE 2: Aborts with an error message if a cyclic dependency is detected.
proc FindIncludeDep { filename } {
    global ArrInclDep INCLUDEDIRS PLATFORM
    set filename [string trim $filename]
    if [info exists ArrInclDep($filename)] { return $ArrInclDep($filename) }
    if [info exists ArrInclDep($filename,WoRkInG)] {
	puts stderr "FATAL ERROR: Cyclic #include dependency detected \
		on file $filename"
	exit 1
    }
    set worklist {}     ;# Dep lists may be empty!
    set ArrInclDep($filename,WoRkInG) {} ;# Guard against cyclic dependencies
    # Ought to use file join
    if {[string compare [file dirname $filename] .]} {
      set pathname $filename
    } else {
      if {[string compare $PLATFORM windows]==0} {
        set pathname [file join [file dirname \
                           [info script]] [file tail $filename]]
      } else {
        set pathname [file dirname [info script]]/[file tail $filename]
      }               
    }
    if [catch {open $pathname "r"} fileid] {
      puts stderr "ERROR: Can't open #include dependency \"$pathname\""
      flush stderr   ;# Safety
      return [set ArrInclDep($filename) {}]
    }
    while { [gets $fileid line] >= 0 } {
	if [regexp "^\[ \t\]*#include\[ \t\]+\"(\[^\"\]+)\"" \
		$line match newfile] {
	    set newfile [string trim $newfile]
            # Search for newfile in the same directory as filename,
            # and also in INCLUDEDIRS
            set newpath [file dirname $filename]/$newfile
            if {![file readable $newpath]} {
              foreach dir $INCLUDEDIRS {
                # Ought to use file join
                if {[string compare $PLATFORM windows]==0} {
                  set newpath [file join $dir $newfile]
                } else {
                  set newpath $dir/$newfile
                }               
                if [file readable $newpath] {
                  break
                }
              }
              if {![file readable $newpath]} {
                puts stderr "ERROR: Can't find #include dependency \"$newfile\""
                puts stderr "located in file $filename ."
                puts stderr "Searched in [file dirname $filename], as well as"
                puts stderr "$INCLUDEDIRS ."
                flush stderr   ;# Safety
                return [set ArrInclDep($filename) {}]
              }
            }
	    # See if $newpath is already in the list.  (We don't want
	    # infinite recursion!)
	    if {[string compare $filename $newpath]==0} continue
	    if {[lsearch -exact $worklist $newpath]<0} {
		# $newpath is a new file
		lappend worklist $newpath
		set worklist [concat \
			$worklist [FindIncludeDep $newpath]]
		# Note: ArrInclDep($filename) must not be defined (set)
		# until *after* these recursive calls.
	    }
	} elseif [regexp -nocase \
		"/\\\*\[ \t\]*end\[ \t\]+includes?\[ \t\]*\\\*/" \
		$line] {
	    # The above is a liberal match against lines including
	    # the sequence '/* End includes */'
	    break
	}
    }
    close $fileid
    # Sort & remove dependencies
    set worklist [lsort $worklist]
    set ArrInclDep($filename) {} ;# Allow for empty Dep lists
    set ilast {}
    foreach i $worklist {
	if { [string compare $i $ilast] != 0 } {
	    lappend ArrInclDep($filename) [set ilast $i]
	}
    }
    unset ArrInclDep($filename,WoRkInG)
    return $ArrInclDep($filename)
}


######################################################################
# Build utility proc's
proc Prepend { prefix namelist } {
    set newlist ""
    foreach i $namelist { lappend newlist "$prefix$i" }
    return $newlist
}
proc Postpend { namelist suffix } {
    set newlist ""
    foreach i $namelist { lappend newlist "$i$suffix" }
    return $newlist
}
proc Compile { srcfile outfile } {
    global CCC CFLAGS EXTRACFLAGS CHANGECOUNT OUTFLAG
    set cmd "$CCC [join $CFLAGS] [join $EXTRACFLAGS] $srcfile $OUTFLAG$outfile"
    puts "$cmd"
    set errcode 0
    if [set exitcode [catch {eval exec $cmd} msg]] {
	puts "--- Compile messages ---\n$msg\n------------------------"
	puts "Exit code: $exitcode"
	set errcode 1
    }
    incr CHANGECOUNT
    return $errcode
}
proc Link { objlist outfile } {
    global CCC LFLAGS LIBXTR CHANGECOUNT RM LINKOUTFLAG
    set cmd "$RM $outfile"
    puts "$cmd"
    catch {eval $cmd}
    set cmd "$CCC $objlist [join $LFLAGS] $LINKOUTFLAG$outfile"
    puts "$cmd"
    set errcode 0
    if [catch {eval exec $cmd} msg] {
	puts "---- Link messages -----\n$msg\n------------------------"
    }
    if {![file exists $outfile] && [string compare $LIBXTR ""]!=0} {
	puts "Retrying link with extra libraries: $LIBXTR"
	if [catch {eval exec $cmd $LIBXTR} msg] {
	    puts "---- Link messages -----\n$msg\n------------------------"
	}
    }
    if ![file exists $outfile] {
	flush stdout
	puts stderr "Sorry; Unable to build $outfile"
	set errcode 1
    } else {
	puts "build: \`$outfile\' complete"
    }
    incr CHANGECOUNT
    return $errcode
}
proc DepCheck { targlist deplist } {
# Returns a sublist of targlist that contains all elements of
# targlist that are older than the newest member of deplist
    global PLATFORM
    set deptime -1
    foreach i $deplist {
        # Ought to use file join
        if {![string compare [file dirname $i] .]} {
          if {[string compare $PLATFORM windows]==0} {
            set i [file join [file dirname [info script]] [file tail $i]]
          } else {
            set i [file dirname [info script]]/[file tail $i]
          }
        }
	if [catch {file mtime $i} modtime] {
	    puts "Error: Dependency $i does not exist"
	} else {
	    if ($modtime>$deptime) { set deptime $modtime }
	}
    }
    set buildlist ""
    foreach i $targlist {
	if {[catch {file mtime $i} targtime] || $targtime<$deptime} { 
	    lappend buildlist $i
	}
    }
    return $buildlist
}
proc MakeDirectory {dir} {
  global MKDIR
  if ![file isdirectory $dir] {
    if [file exists $dir] {
	puts stderr "ERROR: Non-directory $dir already exists"
	exit 1
    }
    if [catch {eval $MKDIR $dir} msg] {
	puts stderr "ERROR: Unable to make directory $dir: $msg"
	exit 1
    }
  }
}
proc InstallFiles { filelist directory } {
  global CP
  if [regexp "\[^ \t\n\]" $filelist] {
    set cmd "$CP [join $filelist] $directory"
#    puts "$cmd"
    catch {eval $cmd}
    flush stderr
  }
}
proc InstallDirectoryTree { srcdir destdir } {
  global CP
  puts stderr "\nInstalling $srcdir as $destdir..."
  if {![file exists $srcdir]} {return}
  if {![file isdirectory $srcdir]} {
    puts stderr "ERROR: InstallDirectoryTree needs directories as arguments."
    exit 1
  }
  MakeDirectory $destdir
  set files {}
  set subdirs {}
  foreach item [glob $srcdir/*] {
    if [file isdirectory $item] {
      if [string compare [file tail $item] CVS] { #Skip CVS directories
        lappend subdirs $item
      }
    }
    if [file isfile $item] {
      lappend files $item
    }
  }
  InstallFiles $files $destdir
  foreach dir $subdirs {
    InstallDirectoryTree $dir $destdir/[file tail $dir]
  }
}

proc MakeSourceDistribution { thisdir destdir } {
  global CP
  puts stderr "\nDistributing $thisdir as $destdir..."
  if {![file exists $thisdir]} {return}
  if {![file isdirectory $thisdir]} {
    puts stderr "ERROR: MakeSourceDistribution needs directories as arguments."
    exit 1
  }
  cd $thisdir
  set savedir [pwd]
  set distribscript distrib.tcl
  if {![file exists $distribscript]} {return}
  if {![file isfile $distribscript]} {return}
  if {![file readable $distribscript]} {return}
  set distributionlist {}
  source $distribscript
#  set distributionlist [Prepend $thisdir/ $distributionlist]
  MakeDirectory $destdir
  set files {}
  set subdirs {}
  foreach item $distributionlist {
    if [file isdirectory $item] {
      lappend subdirs $item
      continue
    }
    if [file isfile $item] {
      lappend files $item
      continue
    }
    puts stderr "Ignoring request to distribute $item ."
    puts stderr "No such file or directory."
  }
  puts stderr "Copying $files to $destdir"
  InstallFiles $files $destdir
  foreach dir $subdirs {
    MakeSourceDistribution $dir $destdir/[file tail $dir]
    cd $savedir
  }
  puts stderr "\nDone distributing $thisdir as $destdir..."
}


# Push the values of a list of variable names onto a stack
proc PushVarList {varlist} {
  global build_stack
  Default build_stack {}
  foreach var $varlist {
    global $var
    lappend build_stack [set $var]
  }
  return [llength $build_stack]
}

# Restore the values retrieved from a stack to a list of variables
proc PopVarList {varlist} {
  global build_stack
  set stacktop [llength $build_stack]
  set stackidx [expr $stacktop - [llength $varlist]]
  set stackpiece [lrange $build_stack $stackidx $stacktop]
  set build_stack [lreplace $build_stack $stackidx $stacktop]
  set i 0
  foreach var $varlist {
    global $var
    set $var [lindex $stackpiece $i]
    incr i
  }
  return [llength $build_stack]
}

proc DoBuild {} {
  global stack argv

  # A list of the variables whose values should not be changed by
  # builds in parent or child directories
  set LOCALVARS {PRESUBDIRS POSTSUBDIRS TARGETS SOURCES EXTINCL BINDIR
                 INCLUDEDIRS OTHEROBJECTS EXTRACFLAGS}
  foreach var $LOCALVARS {
    global $var
  }

  Default EXTRACFLAGS {}

  global CHANGECOUNT MV RM OBJEXT CCCEXT EXEEXT PLATFORM
  global INSTALLROOT INSTALLEXECROOT INSTALLBIN INSTALLTCLLIB INSTALLDEMOS
  global INCLUDEDIRECTORYFLAG SOURCEDISTRIBUTIONDIRECTORY

  # Do builds in the PRE- subdirectories
  PushVarList $LOCALVARS
  foreach dir $PRESUBDIRS {
    # Ought to use file join here
    EvaluateScript $dir/build.tcl
  }
  PopVarList $LOCALVARS

  set CHANGECOUNT 0
  
  # Build object module list from SOURCES
  set OBJECTS $SOURCES
  
  # Add proper suffix to object files
  set OBJECTS [Postpend $OBJECTS $OBJEXT]
  set TARGETOBJ [Postpend $TARGETS $OBJEXT]
  set OTHEROBJECTS [Postpend $OTHEROBJECTS $OBJEXT]
  
  # Add executable extension (if any)
  set TARGETEXE [Postpend $TARGETS $EXEEXT]
  
  # Be sure files are called by pathnames relative to cwd so building can be
  # performed from any working directory.
  set thisdir [file dirname [info script]]
  foreach list $LOCALVARS {
    set $list [Prepend $thisdir/ [set $list]]
  }
  
  # Prepend $OBJECTS and $TARGETS with $BINDIR
  set OBJECTS [Prepend $BINDIR/ $OBJECTS]
  set TARGETOBJ [Prepend $BINDIR/ $TARGETOBJ]
  set TARGETEXE [Prepend $BINDIR/ $TARGETEXE]
  
  # Create BINDIR if necessary
  # Only need it if there are things to build
  if [expr [llength $OBJECTS] + [llength $TARGETS]] {
    MakeDirectory $BINDIR
  }
  
  # Check for "clean" request
  if {[string compare $argv "clean"]==0} {
      set files [concat $OBJECTS $TARGETOBJ $TARGETEXE]
      if [string compare $files {}] {
        set cmd "$RM $files"
        puts "$cmd"
        catch {eval $cmd}
        return
      }
  }
  
  # Check for "install" request
  if {[string compare $argv "install"]==0} {
  puts stderr "Installing..."
      MakeDirectory $INSTALLROOT
      MakeDirectory $INSTALLEXECROOT
      MakeDirectory $INSTALLBIN
      MakeDirectory $INSTALLTCLLIB
      MakeDirectory $INSTALLDEMOS
      InstallFiles $TARGETEXE $INSTALLBIN
  # Ought to use file join
      InstallDirectoryTree [file dirname [info script]]/library $INSTALLTCLLIB
      InstallDirectoryTree [file dirname [info script]]/demos $INSTALLDEMOS
  } else {
  
  # Check for "srcdist" request
  if {[string compare $argv "srcdist"]==0} {
    if [info exists SOURCEDISTRIBUTIONDIRECTORY] {
      set thisdir [file dirname [info script]]
      MakeSourceDistribution $thisdir $SOURCEDISTRIBUTIONDIRECTORY
    } else {
      puts stderr "The variable SOURCEDISTRIBUTIONDIRECTORY must be defined"
      puts stderr "before you can build a source distribution."
      puts stderr "Edit [file dirname [info script]]/build/local.tcl"
      flush stderr
      exit 1
    }
    return
  }
  
  # Change C++ src to have proper extension for compiler/linker
  foreach srcfile [concat [Postpend $SOURCES .cc] [Postpend $TARGETS .cc]] {
      if {[file exists $srcfile] && [string compare $CCCEXT .cc]} {
          regsub {.cc$} $srcfile $CCCEXT destfile
          set cmd "$MV $srcfile $destfile"
          puts "$cmd"
          catch {eval $cmd}
      }
  }
  
  # Add proper suffix to SOURCES
  set SOURCES [Postpend $SOURCES $CCCEXT]
  set TARGETSRC [Postpend $TARGETS $CCCEXT]

  foreach dir $INCLUDEDIRS {
    AppendUnique EXTRACFLAGS $INCLUDEDIRECTORYFLAG$dir
  }

  # Compile objects
  set CompilationErrors 0
  foreach i [concat $SOURCES $TARGETSRC] {
      set deplist [concat $i [FindIncludeDep $i] $EXTINCL]
      set objfile "$BINDIR/[file tail [file rootname $i]]$OBJEXT"
      if {[DepCheck $objfile $deplist]!=""} {
  	incr CompilationErrors [Compile $i $objfile]
      }
  }
  
  # Change C++ src back to have distribution extension
  foreach srcfile [concat $SOURCES $TARGETSRC] {
      if {[file exists $srcfile] && [string compare $CCCEXT .cc]} {
          regsub "$CCCEXT$" $srcfile .cc destfile
          set cmd "$MV $srcfile $destfile"
          puts "$cmd"
          catch {eval $cmd}
      }
  }
  
  if {$CompilationErrors!=0} {
      puts "build: $CompilationErrors Compilation Failure(s)"
      exit 1
  }
  
  # Link objects
  set LinkErrors 0
  foreach trg $TARGETEXE {
      set trgobj [file rootname $trg]$OBJEXT
      set linklist [DepCheck $trg [concat $OBJECTS $trgobj $OTHEROBJECTS]]
      if {"$linklist"!=""} {
          incr LinkErrors [Link [concat $OBJECTS $trgobj $OTHEROBJECTS] $trg]
      }
  }      
  if {$LinkErrors!=0} {
      puts "build: $LinkErrors Link Errors"
      exit 2
  }
  
  if {!$CHANGECOUNT && [string compare $TARGETEXE {}]} {
     puts "build: \`$TARGETEXE\' is up to date."
  }

  }
  
  # Do builds in the POST- subdirectories
  PushVarList $LOCALVARS
  foreach dir $POSTSUBDIRS {
    # Ought to use file join here
    EvaluateScript $dir/build.tcl
  }
  PopVarList $LOCALVARS
}
