# FILE: defaults.tcl
#
# This is a tcl script which performs the configuration task for the
# script ../build.tcl .  It sets values of Tcl variables which describe 
# your computing environment.  
#
# This script is not yet very sophisticated.  It is not able to learn 
# everything it needs to know about your computing environment on its 
# own.  It calls on two other scripts ./local.tcl and ./config.tcl, 
# which must contain definitions for variables which this script cannot
# determine on its own.  You will need to provide some of these 
# definitions by editing the files ./local.tcl and ./config.tcl according
# to the instructions in ./config.tcl .  You should not need to edit
# this file (./defaults.tcl).
#
######################################################################
# First, consult any saved definitions which describe the local platform.
# Consult them first so that they take priority.
set localscript [file dirname [info script]]/local.tcl
if [file exists $localscript] {
  EvaluateScript $localscript
}

######################################################################
# Next, consult any definitions in the (edited) configuration file.
EvaluateScript [file dirname [info script]]/config.tcl

######################################################################
# Finally, try to provide default values for everything, to cover any
# variables not defined in one of the above scripts.

# ------------- Platform description 
# Find out what machine we're building on
set platform_vars {PLATFORM MACHINE OS OSVERSION}

if [AllDefined tcl_platform] {
    Default PLATFORM [string tolower [string trim $tcl_platform(platform)]]
    Default MACHINE [string tolower [string trim $tcl_platform(machine)]]
    Default OS [string tolower [string trim $tcl_platform(os)]]
    Default OSVERSION [string tolower [string trim $tcl_platform(osVersion)]]
}
if {![AllDefined $platform_vars]} {
    ReportMissingValues $platform_vars
}
set OSMAJOR $OSVERSION
regexp {([^.]*).*} $OSVERSION match OSMAJOR

# ------------- System-level commands
set system_vars {MKDIR RM MV CP}

if { [string compare $tcl_version 7.6]==0 } {
    Default MKDIR	"file mkdir"
    Default RM		"file delete"
    Default MV		"file rename"
    Default CP		"file copy"
} else {
  if { [string compare $PLATFORM unix]==0 } {
    Default MKDIR	"exec mkdir"
    Default RM		"exec rm"
    Default MV		"exec mv"
    Default CP		"exec cp"
  } else {
# For other platforms, we need to fill in some configuration defaults.
  }
}
if {![AllDefined $system_vars]} {
    ReportMissingValues $system_vars
}

# ------------- Compiler/Linker configuration
set complink_vars {	CCC CCCEXT OBJEXT EXEEXT COMPILEONLYFLAG 
			MACRODEFINEFLAG INCLUDEDIRECTORYFLAG DEBUGFLAG 
			STRICTWARNFLAG OUTFLAG LINKOUTFLAG
			LIBRARYDIRECTORYFLAG LIBRARYFLAG
                  }

# Select default compiler based on platform/system
if { [string compare $OS irix64]==0 } {
    # 64-bit SGI 
    Default CCC     "CC -32"
} elseif { [string compare $OS sunos]==0 && $OSMAJOR<5} {
    # SunOS
    Default CCC "CC"
} elseif { [string compare $OS sunos]==0 && $OSMAJOR>=5} {
    # Solaris
    Default CCC "CC"
} elseif { [string compare $OS linux]==0 } {
    # Linux
    Default CCC "gcc"
} elseif { [string compare $PLATFORM windows]==0 \
        && [string compare $MACHINE intel]==0} {
    # Windows/intel
    Default CCC "cl"
} else {
    Default CCC "gcc"
}

# Collect any recommended CFLAGS
Default CFLAGS {}

switch [lindex $CCC 0] {
  "CC" {Default CCCEXT 			".cc"
	Default OBJEXT 			".o"
	Default EXEEXT			""
	Default COMPILEONLYFLAG		"-c"
	Default MACRODEFINEFLAG		"-D"
	Default INCLUDEDIRECTORYFLAG	"-I"
	Default DEBUGFLAG		"-g"
	Default STRICTWARNFLAG		"+w"
	Default OUTFLAG 		"-o "	;# Trailing space needed
	Default LINKOUTFLAG 		"-o "	;# Trailing space needed
	Default LIBRARYDIRECTORYFLAG	"-L"
	Default LIBRARYFLAG		"-l"
       }
  "gcc" {Default CCCEXT 			".cc"
	Default OBJEXT 			".o"
	Default EXEEXT			""
	Default COMPILEONLYFLAG		"-c"
	Default MACRODEFINEFLAG		"-D"
	Default INCLUDEDIRECTORYFLAG	"-I"
	Default DEBUGFLAG		"-g"
	Default STRICTWARNFLAG		"-Wall"
	AppendUnique CFLAGS		"-O"
# Note: There is a unfortunate bug in some GCC's:
#  There is a library function 'const char *strchr(const char *,int)',
#  and an inlined function 'char *strchr(char *,int)', that calls
#  the former with casts.  Now, if optimization is turned off, these
#  problematic versions of gcc turn off inlining, and as a result
#  calls to 'char *strchr(char *,int)' will fail at the link step
#  because there is no such library routine.  Sigh...  This problem
#  also affects several other string functions.  See the header file
#  /usr/lib/g++-include/std/cstring.h.
	Default OUTFLAG 		"-o "	;# Trailing space needed
	Default LINKOUTFLAG 		"-o "	;# Trailing space needed
	Default LIBRARYDIRECTORYFLAG	"-L"
	Default LIBRARYFLAG		"-l"
       }
  "cl" {Default CCCEXT 			".cpp"
	Default OBJEXT 			".obj"
	Default EXEEXT			".exe"
	Default COMPILEONLYFLAG		"/c"
	Default MACRODEFINEFLAG		"/D"
	Default INCLUDEDIRECTORYFLAG	"/I"
	Default DEBUGFLAG		""
	Default STRICTWARNFLAG		""
	Default OUTFLAG 		"/Fo"		;# *NO* trailing space
	Default LINKOUTFLAG 		"/link /OUT:"	;# *NO* trailing space
	Default LIBRARYDIRECTORYFLAG	""
	Default LIBRARYFLAG		""
       }
}

if {![AllDefined $complink_vars]} {
  ReportMissingValues $complink_vars
}

# ------------- Compiler/Linker options
set option_vars {CFLAGS LFLAGS LIBXTR}

AppendUnique CFLAGS $COMPILEONLYFLAG

# Construct CFLAGS from configuration information
# Inform compiler how to find Tcl/Tk header files
set includedirs {}
foreach dir {TCLINCLUDEDIRECTORY TKINCLUDEDIRECTORY X11INCLUDEDIRECTORY} {
  AppendUniqueExistingReadableDirectory includedirs $dir
}
foreach dir $includedirs {
  AppendUnique CFLAGS $INCLUDEDIRECTORYFLAG$dir
}

# Construct LFLAGS from configuration information
Default LFLAGS {}

if {[string compare $CCC cl]==0} {
  # Need some defaults here
  Default TCLLIBRARYDIRECTORY "/tcl/bin"
  Default TKLIBRARYDIRECTORY "/tcl/bin"
  regsub {[.]} $tcl_version {} tcl_number
  Default TCLLIBRARYFILE tcl$tcl_number.lib
  # Watch out!  This will break for Tcl 8 ! (or Tcl 7.10)
  set tk_number [expr $tcl_number - 34]
  Default TKLIBRARYFILE tk$tk_number.lib
  set tklibrary $TKLIBRARYDIRECTORY/$TKLIBRARYFILE
  regsub -all {/} $tklibrary {\\\\} tklibrary
  AppendUnique LFLAGS $tklibrary
  set tcllibrary $TCLLIBRARYDIRECTORY/$TCLLIBRARYFILE
  regsub -all {/} $tcllibrary {\\\\} tcllibrary
  AppendUnique LFLAGS $tcllibrary
} else {
  # No known counterexamples for these two library names
  # Set defaults here (extension not really used)
  Default X11LIBRARYFILE "libX11.a"
  Default MATHLIBRARYFILE "libm.a"
  # A good guess for Tcl/Tk libraries
  Default TCLLIBRARYFILE "libtcl.a"
  Default TKLIBRARYFILE "libtk.a"
  set libdirs {}
  foreach dir {TCLLIBRARYDIRECTORY TKLIBRARYDIRECTORY X11LIBRARYDIRECTORY} {
    AppendUniqueExistingReadableDirectory libdirs $dir
  }
  foreach dir $libdirs {
    AppendUnique LFLAGS $LIBRARYDIRECTORYFLAG$dir
  }
#
# This order is important!!!!!
#
  foreach lib {TKLIBRARYFILE TCLLIBRARYFILE X11LIBRARYFILE MATHLIBRARYFILE } {
    if [info exists $lib] {
      if [regexp {^lib(.*)[.][^.]*$} [set $lib] match libab] {
        AppendUnique LFLAGS $LIBRARYFLAG$libab
      }
    }
  }
}

# By default, give the -ldl library flag a try
Default LIBXTR "-ldl"

if {![AllDefined $option_vars]} {
  ReportMissingValues $option_vars
}

# ------------- Installation configuration
set install_vars {INSTALLROOT INSTALLEXECROOT INSTALLBIN INSTALLTCLLIB}

# Default installation directories
Default INSTALLROOT "/usr/local/oommf"
Default INSTALLEXECROOT $INSTALLROOT
Default INSTALLBIN "$INSTALLEXECROOT/bin"
Default INSTALLTCLLIB "$INSTALLROOT/lib"
Default INSTALLDEMOS "$INSTALLROOT/demos"

if {![AllDefined $install_vars]} {
  ReportMissingValues $install_vars
}

