/* FILE: display.cc                 -*-Mode: c++-*-
 *
 * C++ code for display windows
 *
 */

#include "display.h"
#include "misc.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

//////////////////////////////////////////////////////////////////////////
// Colormap class
const ClassDoc DisplayColorMapClass::class_doc("DisplayColorMapClass",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","15-Dec-1996");
DisplayColorMapClass::DisplayColorMapClass(void)
{
#define MEMBERNAME "DisplayColorMapClass"
  if(TKCOLORLENGTH!=6)
    FatalError(-1,STDDOC,"Current code only supports TKCOLORLENGTH==6");
  if(COLORMAPSIZE!=256)
    FatalError(-1,STDDOC,"Current code only supports COLORMAPSIZE==256");
  int i,r,g,b;
  for(i=0;i<COLORMAPSIZE;i++) {
    g=0;
    if(i<COLORMAPSIZE/2) { b=0; r=COLORMAPSIZE-2-2*i; }
    else             { r=0; b=2*(i-COLORMAPSIZE/2); }
    if(r>255) r=255; if(g>255) g=255; if(b>255) b=255; // Safety
    sprintf(TkColor[i],"#%02X%02X%02X",r,g,b);
  }
#undef MEMBERNAME
}

//////////////////////////////////////////////////////////////////////////
// CoordinateSystem helper function
INT4m CoordinateAngleDifference(CoordinateSystem cs1,CoordinateSystem cs2,
				double &angle)
{ // Calculates the difference (in degrees, from 0 to 360) of cs1 - cs2, if
  // both are CS_DISPLAY_* coordinates, sets the difference in 'angle' and
  // returns 0;  If cs1 is not a CS_DISPLAY_* c.s., then the return value
  // is 1; if cs2 is not a CS_DISPLAY_* c.s., then the return value is 2.
  // In these latter two cases angle is 0.0 on return.
  double ang1,ang2;
  angle=0.0;
  switch(cs1) {
  case CS_DISPLAY_STANDARD: ang1=  0; break;
  case CS_DISPLAY_ROT_90:   ang1= 90; break;
  case CS_DISPLAY_ROT_180:  ang1=180; break;
  case CS_DISPLAY_ROT_270:  ang1=270; break;
  default: return 1;  // Illegal coordinates
    // break;
  }
  switch(cs2) {
  case CS_DISPLAY_STANDARD: ang2=  0; break;
  case CS_DISPLAY_ROT_90:   ang2= 90; break;
  case CS_DISPLAY_ROT_180:  ang2=180; break;
  case CS_DISPLAY_ROT_270:  ang2=270; break;
  default: return 2;  // Illegal coordinates
    // break;
  }
  angle=ang1-ang2;
  if(angle<0) angle+=360;
  return 0;
}

//////////////////////////////////////////////////////////////////////////
// Display frame class
const ClassDoc DisplayFrame::class_doc("DisplayFrame",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","14-Dec-1996");
// Recommended display window width, height and cellsize, in pixels:
const REAL4m DisplayFrame::default_width=640;
const REAL4m DisplayFrame::default_height=480;
const CoordinateSystem DisplayFrame::default_coords=CS_DISPLAY_STANDARD;
const REAL4m DisplayFrame::default_cellsize=20;
const REAL4m DisplayFrame::default_arrow_ratio=0.75;

INT4m DisplayFrame::ChangeCoordinates(CoordinateSystem new_coords)
{
  if(display_coords==new_coords) return 0;
  ChangeDisplayBoxCoords(display_coords,new_coords);
  ChangeDisplayListCoords(display_coords,new_coords);
  ChangeBoundaryListCoords(display_coords,new_coords);
  display_coords=new_coords;
  return 0;
}

REAL4m DisplayFrame::GetSampleStepSize(void)
{
  REAL4m step_size=cellsize*mesh_per_pixel;
  REAL4m mesh_cellsize=mesh->GetApproximateCellSize();
  if(mesh_cellsize>EPSILON) step_size/=mesh_cellsize;
  return step_size;
}


INT4m DisplayFrame::SetMeshSampleSize(REAL4m step_request)
{ // Fills display_list from *mesh using requested step size.
  // Copies mesh_box to display_box
  // Use step_request=0 to get _all_ points in mesh.
  // NOTE: Changes display_coords to CS_CALCULATION_STANDARD.  It is
  //       the responsibility of the calling routine to save
  //       display_coords, if desired, _before_ calling this routine.
  // NOTE: Samples mesh in terms of *mesh units*, as opposed to
  //       *mesh cells* (cf. SetSampleStepSize in client interface).
  // NOTE: mesh_per_pixel must be set before calling this routine.

#define MEMBERNAME "SetMeshSampleSize"
  if(step_request<0)
    FatalError(-1,STDDOC,"%s: step_request=%f",ErrBadParam,step_request);
  if(mesh==(Mesh *)NULL) FatalError(-1,STDDOC,"Mesh pointer not set.");

  // Save current coordinate system
  CoordinateSystem orig_coords=display_coords;

  // Get mesh bounding box
  mesh->GetRange(mesh_box);

  // Get list of vectors to display
  REAL4m xstep=step_request;
  REAL4m ystep=step_request;
  REAL4m zstep=0.;
  mesh->GetDisplayList(xstep,ystep,zstep,mesh_box,display_list);

  // Setup display for new cellsize
  step_request=mesh_sample_size=MAX(xstep,ystep);
  if(step_request<mesh->GetApproximateCellSize()/2.)
    step_request=mesh->GetApproximateCellSize();
  cellsize=step_request/mesh_per_pixel;
  if(display_list.GetSize()>0) margin=cellsize/2.;
  else                         margin=0.0;
  arrowsize=REAL4m(cellsize*default_arrow_ratio);
  display_box=mesh_box;
  mesh->GetBoundaryList(boundary_list);
  display_coords=CS_CALCULATION_STANDARD; // reset coordinate system

  // Convert coordinates back to original coordinates
  ChangeCoordinates(orig_coords);

  return 0;
#undef MEMBERNAME
}

INT4m DisplayFrame::SetSampleStepSize(REAL4m step_request)
{ 
  // Fills display_list from *mesh using requested step size.
  // Copies mesh_box to display_box
  // Use step_request=0 to get _all_ points in mesh.
  // NOTE: Changes display_coords to CS_CALCULATION_STANDARD.  It is
  //       the responsibility of the calling routine to save
  //       display_coords, if desired, _before_ calling this routine.
  // NOTE: Samples mesh in terms of *mesh cells*, as opposed to
  // *mesh units* (cf. SetMeshSampleSize in internal interface).
#define MEMBERNAME "SetSampleStepSize"
  if(step_request<0)
    FatalError(-1,STDDOC,"%s: step_request=%f",ErrBadParam,step_request);
  if(mesh==(Mesh *)NULL) FatalError(-1,STDDOC,"Mesh pointer not set.");

  // Convert step_request from mesh cell size to mesh units.
  if(mesh->GetApproximateCellSize()>EPSILON)
    step_request*=mesh->GetApproximateCellSize();

  return SetMeshSampleSize(step_request);
#undef MEMBERNAME
}

INT4m DisplayFrame::SetupFrame(REAL4m requested_width,
			       REAL4m requested_height,
			       CoordinateSystem requested_coords)
{ // Adjusts mesh_per_pixel and mesh_sample_size so that the
  // display list + margin just fits inside a box of the requested
  // width and height. (Note: requested_width & requested_height
  // are in _pixels_.)
#define MEMBERNAME "SetupFrame"
  if(mesh==(Mesh *)NULL) FatalError(-1,STDDOC,"Mesh pointer not set.");

  // Calculations will be done in CS_CALCULATION_STANDARD
  // and CS_DISPLAY_STANDARD coordinates, so swap width
  // and height if the import values are for rotated coords.
  if(requested_coords==CS_DISPLAY_ROT_90 
     || requested_coords==CS_DISPLAY_ROT_270) {
    REAL4m temp=requested_width;
    requested_width=requested_height;
    requested_height=temp;
  }

  // Get mesh bounding box and boundary vertex list
  mesh->GetRange(mesh_box);
  Vec3f mesh_diag=mesh_box.GetDiag();

  // Determine whether width or height is the limiting factor
  REAL4m pixel_dim,mesh_dim;
  if(requested_width*mesh_diag.y <= requested_height*mesh_diag.x) {
    // Width is restriction
    pixel_dim=requested_width; mesh_dim=mesh_diag.x;
  }
  else {
    // Height is restriction
    pixel_dim=requested_height; mesh_dim=mesh_diag.y;
  }
  if(pixel_dim<1. || mesh_dim<EPSILON) pixel_dim=1.0;
             // Safety;  Also handles empty mesh

  // Next determine appropriate step_request, in mesh units
  REAL4m step_request=0;
  if(pixel_dim>default_cellsize+1) {
    step_request=
      default_cellsize*mesh_dim/(pixel_dim-default_cellsize);
  } else {
    step_request=mesh_dim; // pixel_dim is too small; default to 1 step
  }
  mesh_sample_size=mesh->GetApproximateCellSize();
  if(step_request<mesh_sample_size) step_request=mesh_sample_size;
  step_request=mesh->GetPreferredStepSize(step_request);

  // Now determine proper mesh-to-pixel ratio to map mesh region
  // + margin into requested width x height box.
  mesh_per_pixel=(mesh_dim+step_request)/pixel_dim;
  if(mesh_per_pixel<EPSILON) mesh_per_pixel=EPSILON; // Safety

  // Fill display_list
  display_coords=CS_CALCULATION_STANDARD;
  SetMeshSampleSize(step_request);
  ChangeCoordinates(requested_coords);
  return 0;
#undef MEMBERNAME
}

void DisplayFrame::ConvertMeshPtToDisplayPt(Vec3f &point,
					    CoordinateSystem new_coords)
{ // This routine assumes point is using the CS_CALCULATION_STANDARD
  // coordinate system, and display is using CS_DISPLAY_* coordinate
  // system specified by *this->display_coords.
  // NOTE: This routine assumes that display_box is already converted to
  //       "new_coords"
#define MEMBERNAME "ConvertMeshPtToDisplayPt"
#if DEBUGLVL > 9
  if(mesh_per_pixel<EPSILON)
    FatalError(-1,STDDOC,"mesh_per_pixel (=%g) not initialized.",
	       mesh_per_pixel);
#endif
  // First convert to CS_DISPLAY_STANDARD coordinate system
  Vec3f mesh_min,display_min,display_max,vtemp;
  mesh_box.GetMinPt(mesh_min);
  display_box.GetExtremes(display_min,display_max);
  point-=mesh_min;
  point*=(1.0/mesh_per_pixel);
  vtemp.x=rint(point.x)+REAL4m(margin)+display_min.x;
  vtemp.y=display_max.y-rint(point.y)-REAL4m(margin);
  vtemp.z=display_min.z+rint(point.z);

  // Next convert from CS_DISPLAY_STANDARD to new_coords
  point.z=vtemp.z; // This axis is fixed
  switch(new_coords) {
  case CS_DISPLAY_STANDARD:
    point.x=vtemp.x; // Nothing to do.
    point.y=vtemp.y;
    break;
  case CS_DISPLAY_ROT_90:
    point.x=vtemp.y;
    point.y=display_max.y-vtemp.x;
    break;
  case CS_DISPLAY_ROT_180:
    point.x=display_max.x-vtemp.x;
    point.y=display_max.y-vtemp.y;
    break;
  case CS_DISPLAY_ROT_270:
    point.x=display_max.x-vtemp.y;
    point.y=vtemp.x;
    break;
  default:
    FatalError(-1,STDDOC,"%s: Illegal new_coords (%d)",
	       ErrProgramming,new_coords);
    break;
  }
#undef MEMBERNAME
}

void DisplayFrame::ChangeDisplayBoxCoords(CoordinateSystem old_coords,
					  CoordinateSystem new_coords)
{ // Changes display_box coordinates
#define MEMBERNAME "ChangeDisplayListCoords"
  if(old_coords==new_coords) return;
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    if(mesh_per_pixel<EPSILON)
      FatalError(-1,STDDOC,"mesh_per_pixel (=%g) not initialized.",
		 mesh_per_pixel);
    Vec3f disp_diag=(1./mesh_per_pixel)*display_box.GetDiag();
    // Introduce margins
    disp_diag+=Vec3f(REAL4m(2*margin),REAL4m(2*margin),0.);
    display_box.Set(Vec3f(0.,0.,0.),disp_diag);
    // NOTE: display_box is zero-based (as opposed to, e.g., "-margin"-based).
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    ChangeDisplayBoxCoords(old_coords,CS_DISPLAY_STANDARD);
    ChangeDisplayBoxCoords(CS_DISPLAY_STANDARD,new_coords);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    double angle_diff;
    if(CoordinateAngleDifference(new_coords,old_coords,angle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    if(angle_diff==90 || angle_diff==270) { // Swap x & y display ranges
      REAL4m temp;
      Vec3f display_min,display_max;
      display_box.GetExtremes(display_min,display_max);
      temp=display_max.x;
      display_max.x=display_max.y;  display_max.y=temp;
      temp=display_min.x;
      display_min.x=display_min.y;      display_min.y=temp;
      // z coordinate is not changed by these rotations
      display_box.Set(display_min,display_max);
    }
  }
#undef MEMBERNAME
}

void DisplayFrame::ChangeDisplayListCoords(CoordinateSystem old_coords,
					   CoordinateSystem new_coords)
{
  // NOTE: You must call ChangeDisplayBoxCoords() *before* calling this
  // routine, as otherwise the display_box won't be set correcty.

#define MEMBERNAME "ChangeDisplayListCoords"
  if(old_coords==new_coords) return;
  LocatedVector *display_vector;
  Vec3f vtemp;
  Vec3f display_min,display_max;
  display_box.GetExtremes(display_min,display_max);
  // Code below may assume display_min == (0,0,0)

  // Branch through cases
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    if(mesh_per_pixel<EPSILON)
      FatalError(-1,STDDOC,"mesh_per_pixel = %g < %g",
		 mesh_per_pixel,EPSILON);
    REAL4m pixel_per_mesh=1.0/mesh_per_pixel;
    Vec3f mesh_base;
    mesh_box.GetExtremes(mesh_base,vtemp);
    Vec3f left_offset,right_offset;
    // Make margins
    left_offset=display_min-Vec3f(REAL4m(margin),REAL4m(margin),0.);
    right_offset=display_max-Vec3f(REAL4m(margin),REAL4m(margin),0.);
    for(display_vector=display_list.GetFirst();
	display_vector!=(LocatedVector *)NULL;
	display_vector=display_list.GetNext()) {
      // The "->location" section of this could be implemented
      // by a call to ConvertMeshPtToDisplayPt, but this was written
      // before that routine was, and this should be faster for
      // a long list.
      vtemp=display_vector->location;
      vtemp-=mesh_base;  vtemp*=pixel_per_mesh;
      display_vector->location.x = vtemp.x-left_offset.x;
      display_vector->location.y = right_offset.y-vtemp.y;
      display_vector->location.z = right_offset.z-vtemp.z;
      display_vector->value.y*=-1;
    }
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    if(Verbosity>=100) {
      NonFatalError(STDDOC,
	    "Requested coordinate transformation (%d->%d) not yet coded.\n"
	    "  Using 2-step transform (through CS_DISPLAY_STANDARD) instead.",
	    int(old_coords),int(new_coords));
    }
    ChangeDisplayBoxCoords(new_coords,CS_DISPLAY_STANDARD);
    ChangeDisplayListCoords(old_coords,CS_DISPLAY_STANDARD);
    ChangeDisplayBoxCoords(CS_DISPLAY_STANDARD,new_coords);
    ChangeDisplayListCoords(CS_DISPLAY_STANDARD,new_coords);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    double dangle_diff;
    if(CoordinateAngleDifference(new_coords,old_coords,dangle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    INT4m angle_diff=INT4m(rint(dangle_diff));
    for(display_vector=display_list.GetFirst();
	display_vector!=(LocatedVector *)NULL;
	display_vector=display_list.GetNext()) {
      Vec3f *loc=&(display_vector->location);
      Vec3f *dir=&(display_vector->value);
      switch(angle_diff) {
      case 90:
	vtemp = *loc;
	loc->x = vtemp.y;
	loc->y = display_max.y - vtemp.x;
	vtemp = *dir;
	dir->x =  vtemp.y;
	dir->y = -vtemp.x;
	break;
      case 180:
	loc->x = display_max.x - loc->x;
	loc->y = display_max.y - loc->y;
	dir->x *= -1;
	dir->y *= -1;
	break;
      case 270:
	vtemp = *loc;
	loc->x = display_max.x - vtemp.y;
	loc->y = vtemp.x;
	vtemp = *dir;
	dir->x = -vtemp.y;
	dir->y =  vtemp.x;
	break;
      default:
	FatalError(-1,STDDOC,"%s: Illegal angle_diff (%d)",
		   ErrProgramming,angle_diff);
	break;
      }
    }
  }
#undef MEMBERNAME
}

void DisplayFrame::ChangeBoundaryListCoords(CoordinateSystem old_coords,
					    CoordinateSystem new_coords)
{ // Convert boundary vertex list
  // NOTE: You must call ChangeDisplayBoxCoords() *before* calling this
  // routine, as otherwise the display_box won't be set correcty.
#define MEMBERNAME "ChangeBoundaryListCoords"
  if(old_coords==new_coords) return;
  Vec3f *bdry;  // Pointer to boundary_list vertices
  // Branch through cases
  if(old_coords==CS_CALCULATION_STANDARD
     && new_coords==CS_DISPLAY_STANDARD) {
    // Convert from mesh coordinates to display coordinates
    for(bdry=boundary_list.GetFirst(); bdry!=(Vec3f *)NULL;
	bdry=boundary_list.GetNext()) {
      ConvertMeshPtToDisplayPt(*bdry,new_coords);
         // This is presumably a short list
    }
  }
  else if(old_coords==CS_CALCULATION_STANDARD
	  || new_coords==CS_CALCULATION_STANDARD) {
    // Do 2-step change through CS_DISPLAY_STANDARD
    ChangeDisplayBoxCoords(new_coords,CS_DISPLAY_STANDARD);
    ChangeBoundaryListCoords(old_coords,CS_DISPLAY_STANDARD);
    ChangeDisplayBoxCoords(CS_DISPLAY_STANDARD,new_coords);
    ChangeBoundaryListCoords(CS_DISPLAY_STANDARD,new_coords);
  }
  else {  // Flipping between 2 "DISPLAY" coordinate systems
    double dangle_diff;
    if(CoordinateAngleDifference(new_coords,old_coords,dangle_diff)!=0)
      FatalError(-1,STDDOC,"Non-comparable coordinate systems (%d & %d)",
		 old_coords,new_coords);
    INT4m angle_diff=INT4m(rint(dangle_diff));
    Vec3f vtemp,display_min,display_max;
    display_box.GetExtremes(display_min,display_max);
    for(bdry=boundary_list.GetFirst(); bdry!=(Vec3f *)NULL;
	bdry=boundary_list.GetNext()) {
      switch(angle_diff) { // Note: bdry->z doesn't change
      case 90:
	vtemp = *bdry;
	bdry->x = vtemp.y;
	bdry->y = display_max.y - vtemp.x;
	break;
      case 180:
	bdry->x = display_max.x - bdry->x;
	bdry->y = display_max.y - bdry->y;
	break;
      case 270:
	vtemp = *bdry;
	bdry->x = display_max.x - vtemp.y;
	bdry->y = vtemp.x;
	break;
      default:
	FatalError(-1,STDDOC,"%s: Illegal angle_diff (%d) (2nd switch())",
		   ErrProgramming,angle_diff);
	break;
      }
    }
  }
#undef MEMBERNAME
}

INT4m DisplayFrame::DrawArrow(const char *canvas_name,
			      const LocatedVector &locvec)
{
#define MEMBERNAME "DrawArrow"
  const Vec3f *postemp;
  const Vec3f *vtemp;
  REAL4m x1,y1,x2,y2,delx,dely,vec_mag,arrow_len,proj_len;
  INT4m r0,c0,r1,c1,r2,c2,ioff,a1,a2,a3;
  INT4m color;
  CHAR *pcolor1,*pcolor2;

  postemp=&(locvec.location);
  vtemp=&(locvec.value);

  // Line components
  delx=vtemp->x*arrowsize;    dely=vtemp->y*arrowsize;
  vec_mag=vtemp->Mag();
  if(vec_mag<SQRT_EPSILON) return TCL_OK;  // This is a zero vector
  arrow_len=vec_mag*arrowsize;
  proj_len=sqrt(delx*delx+dely*dely);

  // Arrowhead components
  a1=(int)rint(proj_len*0.3);
  a2=(int)rint(proj_len*0.4);
  a3=(int)rint(arrow_len*0.15);
  if(a3<5) a1=a2;

  // Attributes
  int linewidth=(int)rint(a3*0.6);
  if(linewidth<1) linewidth=1;
  color=(int)rint(COLORMAPSIZE*((((vtemp->z)/vec_mag)+1.0)/2.));
  if(color<0)                  color=0;
  else if(color>=COLORMAPSIZE) color=COLORMAPSIZE-1;
  pcolor1=colormap.TkColor[color];

  // NOTE: Timing tests on Linux 2.0.24 on a i486-DX2/66 indicate that
  //       a substantial time loss is involved inside sprintf doing
  //       %f conversions.  It is 6 or 7 times faster to use %d's instead
  //      of sprintf(scratch_buf,"%s create line %6.2f %6.2f %6.2f %6.2f"
  //	             "-arrow last -fill %s",canvas_name,x1,y1,x2,y2,
  //                 colormap.TkColor[color]);

  if(proj_len>=arrow_len*0.15) { // Standard arrow
    x1=(postemp->x)-delx/2;   x2=x1+delx;
    r1=int(rint(x1));     r2=int(rint(x2));
    y1=(postemp->y)-dely/2;   y2=y1+dely;
    c1=int(rint(y1));     c2=int(rint(y2));
    sprintf(scratch_buf,
	    "%s create line %d %d %d %d -arrow last -arrowshape {%d %d %d}"
	    "  -width %d -fill %s",canvas_name,r1,c1,r2,c2,a1,a2,a3,
	    linewidth,pcolor1);
  }
  else { // Small projection with undefined direction.  Draw a dot
    pcolor2=colormap.TkColor[COLORMAPSIZE/2];
    r0=(int)rint(postemp->x); c0=(int)rint(postemp->y);
    if(vtemp->z<0) { // Arrow pointing into-plane
      ioff=(int)rint(REAL4m(a3)/2.);
      r1=r0-ioff;  c1=c0-ioff;
      r2=r0+ioff;  c2=c0+ioff;
      sprintf(scratch_buf,"%s create rect %d %d %d %d -fill %s -outline %s",
	      canvas_name,r1,c1,r2,c2,pcolor1,pcolor1);
      // Note: The return quantity of sprintf is OS dependent.
      if(a3>6) {
	char *bufptr=strchr(scratch_buf,'\0');
	ioff=(int)rint(REAL4m(a3)/6.);
	r1=r0-ioff;  c1=c0-ioff;
	r2=r0+ioff;  c2=c0+ioff;
	sprintf(bufptr,"; %s create rect %d %d %d %d -fill %s -width 0",
		canvas_name,r1,c1,r2,c2,pcolor2);
      }
    } else { // Arrow pointing out-of-plane
      ioff=(int)rint(REAL4m(a3)*(SQRT2/2.));
      r1=r0-ioff;  c1=c0-ioff;
      r2=r0+ioff;  c2=c0+ioff;
      sprintf(scratch_buf,
	      "%s create poly %d %d %d %d %d %d %d %d"
	      " -fill %s -outline %s -width %d",
	      canvas_name,r1,c0,r0,c2,r2,c0,r0,c1,
	      pcolor1,pcolor1,linewidth/3);
      // Note: The return quantity of sprintf is OS dependent.
      if(a3>6) {
	char *bufptr=strchr(scratch_buf,'\0');
	sprintf(bufptr,"; %s create line %d %d %d %d -fill %s -width %d"
		"; %s create line %d %d %d %d -fill %s -width %d",
		canvas_name,r1,c0,r2,c0,pcolor2,linewidth/3,
		canvas_name,r0,c2,r0,c1,pcolor2,linewidth/3);
      }
    }
  }
  if(Tcl_Eval(tcl_interp,scratch_buf) != TCL_OK) return TCL_ERROR;
  return TCL_OK;
#undef MEMBERNAME
}

INT4m DisplayFrame::DrawLine(const char *canvas_name,
			     const Vec3f &Pt1,const Vec3f &Pt2)
{
  sprintf(scratch_buf,
	  "%s create line %6.2f %6.2f %6.2f %6.2f",
	  canvas_name,Pt1.x,Pt1.y,Pt2.x,Pt2.y);
  if(Tcl_Eval(tcl_interp,scratch_buf) != TCL_OK) return TCL_ERROR;
  return TCL_OK;
}

INT4m DisplayFrame::Show(const char *canvas_name)
{
#define MEMBERNAME "Show"
  // Initialize canvas
  if(tcl_interp==NULL)
    FatalError(-1,STDDOC,"tcl_interp not initialized.");
  if(mesh==NULL)
    FatalError(-1,STDDOC,"mesh not initialized.");

  // Check that the display_list is in a displayable coordinate system
  if(display_coords!=CS_DISPLAY_STANDARD &&
     display_coords!=CS_DISPLAY_ROT_90   &&
     display_coords!=CS_DISPLAY_ROT_180  &&
     display_coords!=CS_DISPLAY_ROT_270    ) {
    FatalError(-1,STDDOC,
	       "Can't draw display list in current (%d) coordinate system",
	       display_coords);
  }

  // Draw arrows
  LocatedVector *display_vector;
  for(display_vector=display_list.GetFirst();
      display_vector!=(LocatedVector *)NULL;
      display_vector=display_list.GetNext()) {
    if(DrawArrow(canvas_name,*display_vector) != TCL_OK)
      return TCL_ERROR;
  }

  // Draw mesh edge box
  Vec3f *bdry1,*bdry2;
  if((bdry1=boundary_list.GetFirst())!=(Vec3f *)NULL) {
    while((bdry2=boundary_list.GetNext())!=(Vec3f *)NULL) {
      if(DrawLine(canvas_name,*bdry1,*bdry2) != TCL_OK) return TCL_ERROR;
      bdry1=bdry2;
    }
  }

  return TCL_OK;
#undef MEMBERNAME
}


//   SetDisplayBox changes display_box to bb, reducing extents
// as necessary to keep the current aspect ratio.  If any of the
// extents are negative, then those dimensions will be scaled by
// the others.
//   The new values for display_box are copied into bb upon return.
INT4m DisplayFrame::SetDisplayBox(BoundingBox& bb)
{
  REAL4m curwidth =display_box.GetWidth();
  REAL4m curheight=display_box.GetHeight();
  REAL4m curdepth =display_box.GetDepth();
  REAL4m newwidth =rint(bb.GetWidth());
  REAL4m newheight=rint(bb.GetHeight());
  REAL4m newdepth =rint(bb.GetDepth());

  REAL4m xscale=(curwidth>EPSILON  && newwidth>EPSILON
		 ? newwidth/curwidth   : BIG_VALUE);
  REAL4m yscale=(curheight>EPSILON && newheight>EPSILON
		 ? newheight/curheight : BIG_VALUE);
  REAL4m zscale=(curdepth>EPSILON  && newdepth>EPSILON
		 ? newdepth/curdepth   : BIG_VALUE);

  REAL4m scale=0.0;
  if(xscale<yscale) {
    if(xscale<zscale) scale=xscale;
    else              scale=zscale;
  } else {
    if(yscale<zscale) scale=yscale;
    else              scale=zscale;
  }
  //  NOTE: The above used to be
  //             REAL4m scale=MIN(xscale,MIN(yscale,zscale));
  //        but this died with a "Floating point exception" on Linux/AXP
  //        if compiled with -0# where #>=2.  (mjd,97/3/6)

  if(scale<EPSILON || scale>=BIG_VALUE) {
    // Illegal input; return unchanged
    bb=display_box;
    return 1;
  }

  // Rescale mesh-to-pixel ratio & reload display list
  // (which automatically updates display_box, margin,
  // arrowsize and cellsize).
  mesh_per_pixel/=scale;
  SetMeshSampleSize(mesh_sample_size);
  bb=display_box;
  return 0;
}

INT4m DisplayFrame::SetScale(REAL4m newscale)
{
  if(newscale<EPSILON || newscale>1/EPSILON) return 1; // No change
  mesh_per_pixel=1./newscale;
  // Reload display list, which automatically updates display_box,
  // margin, arrowsize and cellsize.
  SetMeshSampleSize(mesh_sample_size);
  return 0;
}

