/* FILE: display.h                 -*-Mode: c++-*-
 *
 * C++ classes for display windows
 *
 */

#ifndef DISPLAY_H
#define DISPLAY_H

#include <tcl.h>
#include <tk.h>
#include "misc.h"
#include "template.h"
#include "tkcmds.h"
#include "vector.h"

/* End includes */     /* Optional directive to build.tcl */

//////////////////////////////////////////////////////////////////////////
// Colormap class
//
#define COLORMAPSIZE 256
#define TKCOLORLENGTH 6
class DisplayColorMapClass {
  static const ClassDoc class_doc;
public:
  char TkColor[COLORMAPSIZE][TKCOLORLENGTH+2]; // Leave room for '#' & '\0'
  DisplayColorMapClass(void);
};

//////////////////////////////////////////////////////////////////////////
// Display frame class
//
//  Restrictions: Assumes square pixels.
//                2D, ignores z.
//
enum CoordinateSystem { // All are righthanded systems
  CS_ILLEGAL=-1,             // Illegal value
  CS_CALCULATION_STANDARD=0, // x to left,  y up,       z out -  mesh units
  CS_DISPLAY_STANDARD=1,     // x to left,  y down,     z in  - pixel units
  CS_DISPLAY_ROT_90=2,       // x down,     y to right, z in  - pixel units
  CS_DISPLAY_ROT_180=3,      // x to right, y up,       z in  - pixel units
  CS_DISPLAY_ROT_270=4       // x up,       y to left,  z in  - pixel units
};

INT4m CoordinateAngleDifference(CoordinateSystem cs1,CoordinateSystem cs2,
				double &angle);
// Calculates the difference (in degrees, from 0 to 360) of cs1 - cs2, if
// both are CS_DISPLAY_* coordinates, sets the difference in 'angle' and
// returns 0;  If cs1 is not a CS_DISPLAY_* c.s., then the return value
// is 1; if cs2 is not a CS_DISPLAY_* c.s., then the return value is 2.
// In these latter two cases angle is 0.0 on return.


class DisplayFrame
{
private:
  static const ClassDoc class_doc;
  static const REAL4m default_width;    // Recommended display window width
  static const REAL4m default_height;   //  height and
  static const CoordinateSystem default_coords;
  static const REAL4m default_cellsize; //  cellsize, in pixels.
  static const REAL4m default_arrow_ratio; // * cellsize => arrowsize
  // NOTE: Tcl/Tk canvas widget uses floating pt. screen coordinates
  CHAR scratch_buf[SCRATCH_BUF_SIZE];
  REAL4m cellsize;  // Requested cellsize, in pixels
  REAL4m margin;    // Margin around display nodes; typically 1/2 cellsize
  REAL4m arrowsize; // Actual arrowsize, in pixels
  REAL4m mesh_per_pixel;  // * pixels => mesh units;   This is guaranteed>0.
  REAL4m mesh_sample_size; // Mesh sample rate, in mesh units
  BoundingBox display_box;  // Display box, pixel units
  BoundingBox mesh_box;     // Display box, mesh units
  List<Vec3f> boundary_list; // List of boundary vertices
  Mesh *mesh;
  void ConvertMeshPtToDisplayPt(Vec3f &point,CoordinateSystem new_coords);
  List<LocatedVector> display_list;
  enum CoordinateSystem display_coords;
  friend int ChangeDisplayRotation(ClientData,Tcl_Interp *interp,
				   int argc,char** argv);
  void ChangeDisplayBoxCoords(CoordinateSystem old_coords,
			      CoordinateSystem new_coords);
  void ChangeDisplayListCoords(CoordinateSystem old_coords,
			       CoordinateSystem new_coords);
  void ChangeBoundaryListCoords(CoordinateSystem old_coords,
				CoordinateSystem new_coords);
   // Note: Call ChangeDisplayBox *before* calling ChangeDisplayListCoords
  //       or ChangeBoundaryListCoords, in order to properly pre-update
  //       display_box.  None of these routines touch display_coords.
  Tcl_Interp* tcl_interp;
  DisplayColorMapClass colormap;
  INT4m DrawArrow(const char *canvas_name,const LocatedVector &locvec);
  INT4m DrawLine(const char *canvas_name,
		 const Vec3f &Pt1,const Vec3f &Pt2);
  DisplayFrame(DisplayFrame&);            // Disable copy constructor
  DisplayFrame& operator=(DisplayFrame&); // and assignment operator
  INT4m SetMeshSampleSize(REAL4m step_request); // Samples mesh in terms
                                                // of *mesh units*;
public:
  DisplayFrame(void) { mesh=(Mesh *)NULL; tcl_interp=(Tcl_Interp*)NULL;
                       cellsize=0; display_coords=CS_DISPLAY_STANDARD; }
  void SetMesh(Mesh *newmesh) { mesh=newmesh; }
  void SetTclInterp(Tcl_Interp *newinterp) { tcl_interp=newinterp; }
  INT4m SetupFrame(REAL4m requested_width=default_width,
		   REAL4m requested_height=default_height,
		   CoordinateSystem requested_coords=default_coords);
  INT4m Show(const char *canvas_name);
  INT4m ChangeCoordinates(CoordinateSystem new_coords);
  CoordinateSystem GetCoordinates() { return display_coords; }
  REAL4m GetSampleStepSize(void);
  INT4m SetSampleStepSize(REAL4m step_request); // Samples mesh in terms
                                                // of mesh cell size;
  REAL4m GetScale(void) const { return 1./mesh_per_pixel; }
  INT4m SetScale(REAL4m newscale);
  BoundingBox GetDisplayBox() const { return display_box; }
  INT4m SetDisplayBox(BoundingBox& bb);
};

#endif // DISPLAY_H




