/* FILE: fileio.cc                 -*-Mode: c++-*-
 *
 * File input/output routines
 *
 */


#include "fileio.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

#define VECDIM 3  // Currently *assume* all vectors are 3D

// Regular 3D grid of Vec3f's
const ClassDoc VioFile::class_doc("VioFile",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","16-Dec-1996");

INT4m VioFile::OpenFileForInput(const char *filename)
{
#define MEMBERNAME "OpenFileForInput"
  CloseFile();
  if((fptr=fopen(filename,"rb"))==(FILE *)NULL) {
    NonFatalError(STDDOC,"%s %s",ErrBadFileOpen,filename);
    return 1;
  }
  return ReadHeader();
#undef MEMBERNAME
}

INT4m VioFile::ReadHeader(void)
{
#define MEMBERNAME "ReadHeader"
  // Init class member variables to 0 (in case of errors).
  width=height=depth=0;

  if(fptr==(FILE *)NULL)
    { NonFatalError(STDDOC,"File stream not initialized"); return 1; }

  struct VioFileHeader {
    CHAR    Magic[8];
    REAL8   EndianTest;
    INT4    Nx,Ny,Nz,dimen,dataoffset;
    CHAR    Reserved[36];
    CHAR    Note[448];
  } head;

  const char   VecFileMagic02[8] = "VecFil\x02";
  const double VecFileEndianTest = 1.234;
  // The bit pattern for 1.234 on a little endian machine (e.g.,
  // Intel's x86 series) should be 58 39 B4 C8 76 BE F3 3F, while
  // on a big endian machine it is 3F F3 BE 76 C8 B4 39 58.  Examples
  // of big endian machines include ones using the MIPS R4000 and R8000
  // series chips, for example SGI's.

  // Read & test magic cookie
  if(fread(head.Magic,1,sizeof(head.Magic),fptr)!=sizeof(head.Magic))
    { NonFatalError(STDDOC,ErrInputEOF); return 1; }
  if(strcmp(head.Magic,VecFileMagic02)!=0)
    { NonFatalError(STDDOC,ErrBadFileHeader); return 1; }

  // Read rest of header
  if(fread((char *)&head.EndianTest,sizeof(head.EndianTest),1,fptr)!=1)
    { NonFatalError(STDDOC,ErrInputEOF); return 1; }
  if(fread((char *)&head.Nx,sizeof(head.Nx),1,fptr)!=1)
    { NonFatalError(STDDOC,ErrInputEOF); return 1; }
  if(fread((char *)&head.Ny,sizeof(head.Ny),1,fptr)!=1)
    { NonFatalError(STDDOC,ErrInputEOF); return 1; }
  if(fread((char *)&head.Nz,sizeof(head.Nz),1,fptr)!=1)
    { NonFatalError(STDDOC,ErrInputEOF); return 1; }
  if(fread((char *)&head.dimen,sizeof(head.dimen),1,fptr)!=1)
    { NonFatalError(STDDOC,ErrInputEOF); return 1; }
  if(fread((char *)&head.dataoffset,sizeof(head.dataoffset),1,fptr)!=1)
    { NonFatalError(STDDOC,ErrInputEOF); return 1; }
  if(fread(head.Reserved,1,sizeof(head.Reserved),fptr)
     !=sizeof(head.Reserved))
    { NonFatalError(STDDOC,ErrInputEOF); return 1; }
  if(fread(head.Note,1,sizeof(head.Note),fptr)!=sizeof(head.Note))
    { NonFatalError(STDDOC,ErrInputEOF); return 1; }

  // Determine byte order
  if(head.EndianTest==VecFileEndianTest) swap_input_bytes=0;
  else if(flip(head.EndianTest)==VecFileEndianTest) swap_input_bytes=1;
  else { NonFatalError(STDDOC,"Unknown binary byte order"); return 1; }
  if(swap_input_bytes) {
    if(Verbosity>=100) Message("Reversing binary byte order on input\n");
    head.Nx=flip(head.Nx);
    head.Ny=flip(head.Ny);
    head.Nz=flip(head.Nz);
    head.dimen=flip(head.dimen);
    head.dataoffset=flip(head.dataoffset);
  }

  // Check for nonsensical input
  if(head.dataoffset!=sizeof(VioFileHeader))
    { NonFatalError(STDDOC,ErrBadFileHeader); return 1; }
  if(strlen(head.Note)>=sizeof(head.Note))
    { NonFatalError(STDDOC,ErrBadFileHeader); return 1; }
  if(head.Nx<1 || head.Ny<1 || head.Nz<1 || head.dimen<1)
    { NonFatalError(STDDOC,ErrBadFileHeader); return 1; }
  if(head.dimen!=VECDIM)
    { NonFatalError(STDDOC,"Only %dD vectors (not %dD) currently supported",
		    VECDIM,head.dimen); return 1; }

  // Set class member variables
  width=head.Nx;  height=head.Nz; depth=head.Ny;
  // vio files use x & z in plane, y pointing into plane

  return 0;
#undef MEMBERNAME
}

INT4m VioFile::FillArray(GridVec3f &grid)
{ // Reads magnetization vector array.
  // Assumes fptr is correctly positioned.
#define MEMBERNAME "FillArray"
  if(fptr==(FILE *)NULL)
    FatalError(-1,STDDOC,"File stream not initialized");
  INT4m grid_width,grid_height,grid_depth;
  grid.GetSize(grid_width,grid_height,grid_depth);
  if(grid_width!=width || grid_height!=height || grid_depth!=depth)
    FatalError(-1,STDDOC,"Grid not properly sized for incoming data."
	       "  Try using VioFile::GetDimensions().");

  // Read in data, 1 column at a time (vio file is stored columnwise)
  REAL8 *rowbuf=new REAL8[height*VECDIM];
  if(rowbuf==NULL) FatalError(-1,STDDOC,ErrNoMem);
  for(INT4m k=0;k<depth;k++) for(INT4m i=0;i<width;i++) {
    if(fread((char *)rowbuf,sizeof(REAL8)*VECDIM,height,fptr)
       !=(unsigned int)height)
      FatalError(-1,STDDOC,"%s (k=%d, i=%d)",ErrInputEOF,k,i);
    if(swap_input_bytes) swap_words8(rowbuf,height*VECDIM);
#if VECDIM != 3
#error "Code in VioFile::FillArray assumes 3D vector input"
#endif
    for(INT4m j=0;j<height;j++) {
      grid(i,j,k)=Vec3f( REAL4(rowbuf[3*j]),
			 REAL4(rowbuf[3*j+2]),
                        -REAL4(rowbuf[3*j+1]));
      // vio file   uses x => left, z => up, y => into plane,
      // while grid uses x => left, y => up, z => out of plane.
    }
  }
  delete[] rowbuf;
  return 0;
#undef MEMBERNAME
}
