/* FILE: misc.cc                 -*-Mode: c++-*-
 *
 * Useful miscellaneous stuff.
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "misc.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

int Verbosity=10; // Default value for Verbosity
#define BigBufSize   4096   // Size used for large (normally char) buffers
#define MedBufSize   1024   // "" medium sized buffers
#define SmallBufSize  256   // "" small buffers

const char* ErrArrOutOfBounds = "Array index out of bounds";
const char* ErrBadFile        = "File error";
const char* ErrBadFileHeader  = "Bad file header";
const char* ErrBadFileOpen    = "Unable to open file";
const char* ErrInputEOF       = "Premature End-Of-File on input";
const char* ErrOutputEOF      = "Premature End-Of-File on output";
const char* ErrBadParam       = "Bad parameter";
const char* ErrDispInit       = "Unable to initialize display";
const char* ErrMisc           = "Miscellaneous error";
const char* ErrNoMem          = "Insufficent working memory";
const char* ErrOverflow       = "Arithmetic overflow";
const char* ErrProgramming    = "Programming error";

ClassDoc::ClassDoc(const CHAR *new_classname,const CHAR *new_maintainer,
		   const CHAR *new_revision,const CHAR *new_revdate)
{
  classname  = new_classname;
  maintainer = new_maintainer;
  revision   = new_revision;
  revdate    = new_revdate;
}

//////////////////////////////////////////////////////////////////////////
///////////////////////////// ERROR HANDLERS /////////////////////////////
//////////////////////////////////////////////////////////////////////////
//
// See also TkMessage() in tkcmds.cc

static char ErrorBuf[BigBufSize];

// Things worth dying over...
void FatalError(INT4m errorcode,
		const CHAR *membername,
		const ClassDoc &class_doc,
		const CHAR *filename,INT4m lineno,
		const CHAR *fmt,...)
{
  va_list arg_ptr;
  char *cptr;
  sprintf(ErrorBuf,
	  "Fatal error occurred in %s::%s (File %s, line %d)\n"
	  "Class revision/date: %s/%s\n"
	  "Class maintainer: %s\nERROR: ",
	  class_doc.classname,membername,filename,lineno,
	  class_doc.revision,class_doc.revdate,
	  class_doc.maintainer);
  cptr=strchr(ErrorBuf,'\0'); // Note: Return type of sprintf is OS dependent
  va_start(arg_ptr,fmt);
  vsprintf(cptr,fmt,arg_ptr);
  va_end(arg_ptr);

  ErrorBuf[sizeof(ErrorBuf)-1]='\0'; // A wee bit of safety
  TkMessage(MSG_ERROR,ErrorBuf,errorcode);

  // We shouldn't get to here, but just in case...
  exit(1);
}

// Fatal Error handler for non-class functions
void PlainError(INT4m errorcode,const CHAR *fmt,...)
{
  va_list arg_ptr;
  va_start(arg_ptr,fmt);
  vsprintf(ErrorBuf,fmt,arg_ptr);
  va_end(arg_ptr);

  ErrorBuf[sizeof(ErrorBuf)-1]='\0'; // A wee bit of safety
  TkMessage(MSG_ERROR,ErrorBuf,errorcode);

  // We shouldn't get to here, but just in case...
  exit(1);
}


// Warnings
void NonFatalError(const CHAR *membername,
		   const ClassDoc &class_doc,
		   const CHAR *filename,INT4m lineno,
		   const CHAR *fmt,...)
{
  va_list arg_ptr;
  char *cptr;
  sprintf(ErrorBuf,
	  "Nonfatal error occurred in %s::%s (File %s, line %d)\n"
	  "Class revision/date: %s/%s\n"
	  "Class maintainer: %s\nWARNING: ",
	  class_doc.classname,membername,filename,lineno,
	  class_doc.revision,class_doc.revdate,
	  class_doc.maintainer);
  cptr=strchr(ErrorBuf,'\0'); // Note: Return type of sprintf is OS dependent
  va_start(arg_ptr,fmt);
  vsprintf(cptr,fmt,arg_ptr);
  va_end(arg_ptr);

  ErrorBuf[sizeof(ErrorBuf)-1]='\0'; // A wee bit of safety
  TkMessage(MSG_WARNING,ErrorBuf,0);
}

// Nonfatal error handler for non-class functions
void PlainWarning(const CHAR *fmt,...)
{
  va_list arg_ptr;
  va_start(arg_ptr,fmt);
  vsprintf(ErrorBuf,fmt,arg_ptr);
  va_end(arg_ptr);

  ErrorBuf[sizeof(ErrorBuf)-1]='\0'; // A wee bit of safety
  TkMessage(MSG_WARNING,ErrorBuf,0);
}

// Informational messages
int Message(const CHAR *fmt,...)
{
  va_list arg_ptr;
  va_start(arg_ptr,fmt);
  vsprintf(ErrorBuf,fmt,arg_ptr);
  va_end(arg_ptr);

  ErrorBuf[sizeof(ErrorBuf)-1]='\0'; // A wee bit of safety
  return TkMessage(MSG_INFO,ErrorBuf,0);
}

//////////////////////////////////////////////////////////////////////////
// Byte reversal routines;  Converts between big-endian
// and little-endian representations.
void swap_bytes(void *buf,INT4m buflen)
{
  swab((char *)buf,(char *)buf,buflen);
}

void swap_words4(void *buf,INT4m count)
{
  UINT4 *bufptr;
  bufptr=(UINT4 *)buf;
  for(;count>0;++bufptr,--count) (*bufptr)=flip(*bufptr);  
}

void swap_words8(void *buf,INT4m count)
{
  REAL8 *bufptr;  // Type is irrelevant, as long as it is 8 bytes wide
  bufptr=(REAL8 *)buf;
  for(;count>0;++bufptr,--count) (*bufptr)=flip(*bufptr);  
}


//////////////////////////////////////////////////////////////////////////
// String class
//
const ClassDoc String::class_doc("String",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","18-Mar-1997");

String::String()
{
#define MEMBERNAME "String"
  str=new char[1];
  if(str==0) FatalError(-1,STDDOC,"No memory");
  str[0]='\0';
#undef MEMBERNAME
}

String::String(const char *cptr)
{
#define MEMBERNAME "String(const char *cptr)"
  if(cptr==NULL) str=new char[1];
  else str=new char[strlen(cptr)+1];
  if(str==0) FatalError(-1,STDDOC,"No memory");
  if(cptr==NULL) str[0]='\0';
  else strcpy(str,cptr);
#undef MEMBERNAME
}

String::String(const String &rhs)
{
#define MEMBERNAME "String(const String &)"
  str=new char[rhs.Length()+1];
  if(str==0) FatalError(-1,STDDOC,"No memory");
  strcpy(str,rhs.str);
#undef MEMBERNAME
}

String::~String()
{
#define MEMBERNAME "~String"
  delete[] str;
#undef MEMBERNAME
}

String& String::operator=(const String &rhs)
{ // Assignment operator
#define MEMBERNAME "operator=(const String &)"
  if(&rhs==this) return *this;  // Nothing to do!
  delete[] str;
  str=new char[rhs.Length()+1];
  if(str==0) FatalError(-1,STDDOC,"No memory");
  strcpy(str,rhs.str);
  return *this;
#undef MEMBERNAME
}

String& String::operator=(const char *cptr)
{ // Assignment operator
#define MEMBERNAME "operator=(const char *)"
  if(cptr==NULL) FatalError(-1,STDDOC,"Null pointer passed (cptr)");
  if(cptr==str) return *this;  // Nothing to do!
  delete[] str;
  str=new char[strlen(cptr)+1];
  if(str==0) FatalError(-1,STDDOC,"No memory");
  strcpy(str,cptr);
  return *this;
#undef MEMBERNAME
}

INT4m StringCompare(const String &a,const String &b)
{ // Replicates strcmp
  return strcmp(a.str,b.str);
}

INT4m StringSpan(const String &a,const char *b)
{ // Replicates strspn
  return strspn(a.str,b);
}

// Append functions; grows str as needed.  The char version appends
// up to the first '\0' character or maxlen, which ever happens first.
// Both return *this.
String& String::Append(const String& appendstr)
{
  if(appendstr.str[0]=='\0') return *this; // Nothing to do.
  int len1=Length();  int len2=appendstr.Length();
  int len3=len1+len2;
  char *newstr=new char[len3+1];
  memcpy(newstr,str,len1);
  memcpy(newstr+len1,appendstr,len2);
  newstr[len3]='\0';
  delete[] str;
  str=newstr;
  return *this;
}

String& String::Append(const char *appendarr,int maxlen)
{
  // Determine length of string to append
  int len2=0;
  while(len2<maxlen && appendarr[len2]!='\0') len2++;
  if(len2==0) return *this; // Nothing to do.
  int len1=Length();
  int len3=len1+len2;
  char *newstr=new char[len3+1];
  memcpy(newstr,str,len1);
  memcpy(newstr+len1,appendarr,len2);
  newstr[len3]='\0';
  delete[] str;
  str=newstr;
  return *this;
}

// Function to trim leading & trailing whitespace (as determined by
// 'isspace' function) from string.
// NOTE: This routine currently operates "in-place", and
String& String::Trim()
{
  char *cptr1,*cptr2,*cptr3;
  cptr1=str;
  while(*cptr1!='\0' && isspace(*cptr1)) cptr1++;
  if(*cptr1=='\0') {
    str[0]='\0'; // Blank string
    return *this;
  }
  // From this point down, we know *cptr1 is non-space
  cptr2=strchr(cptr1,'\0')-1;
  while(isspace(*cptr2)) cptr2--;
  *(++cptr2)='\0'; // cptr2 points to first whitespace character in trailer

  if(cptr1!=str) {
    // Down copy string between cptr1 & cptr2
    cptr3=str;
    while(cptr1<cptr2) {
      *(cptr3++) = *(cptr1++);
    }
  }
  return *this;
}


