/* FILE: template.h                 -*-Mode: c++-*-
 *
 * C++ template declarations
 *
 */

#ifndef TEMPLATE_H
#define TEMPLATE_H

#include "misc.h"

/* End includes */     /* Optional directive to build.tcl */

//////////////////////////////////////////////////////////////////////////
// 3D array template
template<class T> class Array3D {
private:
  const ClassDoc class_doc; // Static data member templates not implemented
                            // by gnu C++ compiler.
  INT4m dim1,dim2,dim3;  // I would *like* these private
  T ***arr;              // I would *like* this private
public:
  INT4m Copy(Array3D<T> &src);
  Array3D();
  Array3D(Array3D<T> &src); // Copy constructor
  ~Array3D() { Deallocate(); }
  INT4m Allocate(INT4m newdim1,INT4m newdim2,INT4m newdim3);
  void Deallocate();
  void Swap(Array3D<T> &other); // Interchanges *this & other
  INT4m GetSize(INT4m &d1,INT4m &d2,INT4m &d3) const
     { d1=dim1; d2=dim2; d3=dim3; return d1*d2*d3; }
  INT4m GetSize() const { return dim1*dim2*dim3; }
  inline T& operator()(INT4m i1,INT4m i2,INT4m i3);
  inline const T& operator()(INT4m i1,INT4m i2,INT4m i3) const;
};

template<class T> Array3D<T>::Array3D(void) : 
  class_doc("Array3D","Michael J. Donahue (michael.donahue@nist.gov)",
	    "1.0.0","12-Dec-1996")
     // Static data member templates not implemented by gnu C++ v2.7.1.
{
  dim1=dim2=dim3=0; arr=(T***)NULL;
}

template<class T> Array3D<T>::Array3D(Array3D<T> &src) :
  class_doc("Array3D","Michael J. Donahue (michael.donahue@nist.gov)",
	    "1.0.0","12-Dec-1996")
     // Static data member templates not implemented by gnu C++ v2.7.1.
{ // Copy constructor
#define MEMBERNAME "Array3D(Array3D<T> &)"
  dim1=dim2=dim3=0; arr=(T***)NULL;

  Copy(src);
#undef MEMBERNAME
}

template<class T> INT4m Array3D<T>::Copy(Array3D<T> &src)
{ 
#define MEMBERNAME "Copy(Array3D<T> &)"
  if(&src==this) return GetSize(); // Nothing to do!
  Allocate(src.dim1,src.dim2,src.dim3);

  // Copy data
  INT4m total_size=GetSize();
  T *srcdata=&(src.arr[0][0][0]);
  T *destdata=&(arr[0][0][0]);
  for(INT4m i=0;i<total_size;i++) destdata[i]=srcdata[i];
  return total_size;
#undef MEMBERNAME
}

template<class T> void Array3D<T>::Deallocate(void)
{
#define MEMBERNAME "Deallocate"
  if(arr!=(T***)NULL) {
    delete[] arr[0][0];
    delete[] arr[0];
    delete[] arr;
  }
  arr=(T***)NULL;
  dim1=dim2=dim3=0;
#undef MEMBERNAME
}

template<class T> INT4m Array3D<T>::Allocate(INT4m newdim1,INT4m newdim2,
					     INT4m newdim3)
{ // Insufficient memory is currently a fatal error.
  //  This may change (but requires careful deallocation). -mjd 12/Dec/96
#define MEMBERNAME "Allocate"
  // Check parameters, deallocate old memory and/or handle special cases.
  if(newdim1<0 || newdim2<0 || newdim3<0) FatalError(-1,STDDOC,ErrBadParam);
  if(dim1==newdim1 && dim2==newdim2 && dim3==newdim3) return GetSize();
  Deallocate();
  if(newdim1==0 || newdim2==0 || newdim3==0) return 0;

  INT4m i1,i2,total_size;
  if((total_size=newdim1*newdim2*newdim3)<1) // Partial check
    FatalError(-1,STDDOC,ErrOverflow);

  // Allocate memory for plane pointers
  if(!(arr=new T**[newdim1])) FatalError(-1,STDDOC,ErrNoMem);

  // Allocate memory for row pointers & initialize plane pointers
  if(!(arr[0]=new T*[newdim1*newdim2])) FatalError(-1,STDDOC,ErrNoMem);
  for(i1=1;i1<newdim1;i1++) arr[i1]=arr[0]+i1*newdim2;

  // Allocate memory for individual cells & initialize row pointers
  if(!(arr[0][0]=new T[total_size])) FatalError(-1,STDDOC,ErrNoMem);
  for(i1=0;i1<newdim1;i1++) for(i2=0;i2<newdim2;i2++) {
    arr[i1][i2]=arr[0][0]+(i1*newdim2+i2)*newdim3;
  }

  dim1=newdim1; dim2=newdim2; dim3=newdim3;
  return total_size;
#undef MEMBERNAME
}

template<class T> void Array3D<T>::Swap(Array3D<T> &other)
{
#define MEMBERNAME "Swap"
  INT4m tdim1,tdim2,tdim3; T*** tarr;
  tdim1=dim1;       tdim2=dim2;       tdim3=dim3;       tarr=arr;
  dim1=other.dim1;  dim2=other.dim2;  dim3=other.dim3;  arr=other.arr;
  other.dim1=tdim1; other.dim2=tdim2; other.dim3=tdim3; other.arr=tarr;
#undef MEMBERNAME
}

template<class T> T& Array3D<T>::operator()(INT4m i1,INT4m i2,INT4m i3)
{
#define MEMBERNAME "operator()(INT4m i1,INT4m i2,INT4m i3)"
#if DEBUGLVL >= 10
  if(i1<0 || i1>=dim1 || i2<0 || i2>=dim2 || i3<0 || i3>=dim3)
    FatalError(-1,STDDOC,ErrArrOutOfBounds);
#endif
  return arr[i1][i2][i3];
#undef MEMBERNAME
}

template<class T> const T&
   Array3D<T>::operator()(INT4m i1,INT4m i2,INT4m i3) const
{
#define MEMBERNAME "operator() const (INT4m i1,INT4m i2,INT4m i3)"
#if DEBUGLVL >= 10
  if(i1<0 || i1>=dim1 || i2<0 || i2>=dim2 || i3<0 || i3>=dim3)
    FatalError(-1,STDDOC,ErrArrOutOfBounds);
#endif
  return arr[i1][i2][i3];
#undef MEMBERNAME
}

//////////////////////////////////////////////////////////////////////////
// List template
template<class T> class List {
private:
  const ClassDoc class_doc; // Static data member templates not implemented
                            // by gnu C++ compiler.
  INT4m extend_length;     // Number of elements to add to List when
                           // it is necessary to increase size of List.
  T *arr;                  // List array
  INT4m length,allocated_length; // Length used & allocated for arr
  INT4m index;                   // Next access element
  void Extend();                 // Add BlockSize elts. to arr
  void Deallocate();             // Release arr.
  void Init();
public:
  INT4m Copy(const List<T> &src);
  List(): class_doc("List","Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","13-Dec-1996")
       // Static data member templates not implemented by gnu C++ v2.7.1.
  { Init(); }

  List(const List<T> &src); // Copy constructor
  ~List() { Deallocate(); }
  void Free(void) { Deallocate(); }
  INT4m GetSize() { return length; }
  T *GetFirst(){ index=0; return GetNext(); }
  T *GetNext() { if(index>=length) return NULL; return arr+(index++); }
  void Clear(void) { index=0; length=0; }
  INT4m SetFirst(const T &elt); // Copies elt & sets length to 1.
  INT4m Append(const T &elt);   // Indexes off of "length"
  List<T> &operator=(const List<T> &other) { Copy(other); return *this; }
};

template<class T> void List<T>::Init(void)
{
#define MEMBERNAME "Init()"
  length=allocated_length=0; arr=(T*)NULL;
  index=0;
  extend_length=MEMBLOCKSIZE/sizeof(T);
#undef MEMBERNAME
}

template<class T> void List<T>::Deallocate(void)
{
#define MEMBERNAME "Deallocate"
  if(arr!=(T*)NULL) delete[] arr;
  Init();
#undef MEMBERNAME
}

template<class T> INT4m List<T>::Copy(const List<T> &src)
{
#define MEMBERNAME "Copy(List<T> &)"
  if(&src==this) return 0;  // Nothing to do!
  if(allocated_length!=src.allocated_length) {
    Deallocate();
    if(!(arr=new T[src.allocated_length]))
      FatalError(-1,STDDOC,ErrNoMem);
    allocated_length=src.allocated_length;
  }
  index=0; length=0;
  while(length<src.length) {
    arr[length]=src.arr[length];  // This is SUPPOSED to be an appropriate
                                  // copy operator!!!
    length++;
  }
  return 0;
#undef MEMBERNAME
}

template<class T> void List<T>::Extend(void)
{
#define MEMBERNAME "Extend"
  T *tarr;
  if(!(tarr=new T[allocated_length+extend_length]))
    FatalError(-1,STDDOC,ErrNoMem);
  for(INT4m i=0;i<length;i++) tarr[i]=arr[i];
  delete[] arr;
  arr=tarr;
  allocated_length+=extend_length;
#undef MEMBERNAME
}

#ifdef REALLY_DUMB
template<class T> List<T>::List(void) : 
  class_doc("List","Michael J. Donahue (michael.donahue@nist.gov)",
	    "1.0.0","13-Dec-1996")
     // Static data member templates not implemented by gnu C++ v2.7.1.
{
#define MEMBERNAME "List()"
  Init();
#undef MEMBERNAME
}
#endif // REALLY_DUMB

template<class T> List<T>::List(const List<T> &src) :
  class_doc("Array3D","Michael J. Donahue (michael.donahue@nist.gov)",
	    "1.0.0","12-Dec-1996")
     // Static data member templates not implemented by gnu C++ v2.7.1.
{ // Copy constructor
#define MEMBERNAME "List(List<T> &)"
  Init();
  Copy(src);
#undef MEMBERNAME
}

template<class T> INT4m List<T>::Append(const T &elt)
{
#define MEMBERNAME "Append"
  if(length>=allocated_length) {
    Extend();
    if(length>=allocated_length) FatalError(-1,STDDOC,ErrProgramming);
  }
  arr[length++]=elt;
  return 0;
#undef MEMBERNAME
}

template<class T> INT4m List<T>::SetFirst(const T &elt)
{
#define MEMBERNAME "SetFirst"
  Clear();
  return Append(elt);
#undef MEMBERNAME
}

#endif // TEMPLATE_H
