#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "vecfile.h"
#include "fileio.h"  // Obsoleted *.vio input

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

///////////////////////////////////////////////////////////////////////////
// VectorFieldFileInputID Class

const ClassDoc VectorFieldFileInputID::class_doc("VectorFieldFileInputID",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","18-Mar-1997");


VectorFieldFileInputID::VectorFieldFileInputID(
		 const char *new_filetype,
		 BOOL (*newpIsType)(FILE *),
		 VectorFieldFileInput* (*newpMakeNew)(const char *,FILE *),
		 BOOL (*pRegister)(const VectorFieldFileInputID &))
  : filetype(new_filetype)
{
#define MEMBERNAME "VectorFieldFileInputID"
  pIsType=newpIsType;
  pMakeNew=newpMakeNew;
  if(pRegister!=NULL)  (*pRegister)(*this);
#undef MEMBERNAME
}

void VectorFieldFileInputID::Dup(const VectorFieldFileInputID &rhs)
{ // Assignment operator
#define MEMBERNAME "Dup"
  filetype=rhs.filetype;
  pIsType=rhs.pIsType;
  pMakeNew=rhs.pMakeNew;
#undef MEMBERNAME
}

VectorFieldFileInputID& VectorFieldFileInputID::operator=(
			   const VectorFieldFileInputID &rhs)
{ // Assignment operator
  Dup(rhs);
  return *this;
}

VectorFieldFileInputID::VectorFieldFileInputID(const VectorFieldFileInputID
					       &copyobj) // Copy constructor
{
  Dup(copyobj);
}

VectorFieldFileInputID::VectorFieldFileInputID()
{ // Default constructor; Needed by List<T>
#define MEMBERNAME "VectorFieldFileInputID()"
  pIsType=NULL;
  pMakeNew=NULL;
#undef MEMBERNAME
}

VectorFieldFileInputID::~VectorFieldFileInputID()
{ // Dummy function for now
#define MEMBERNAME "~VectorFieldFileInputID"
#undef MEMBERNAME
}

///////////////////////////////////////////////////////////////////////////
// VectorFieldFileInput Class

const ClassDoc VectorFieldFileInput::class_doc("VectorFieldFileInput",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","17-Mar-1997");

List<VectorFieldFileInputID> VectorFieldFileInput::child_list;

VectorFieldFileInput::VectorFieldFileInput(const char *input_filename,
					   FILE *new_infile)
  : filename(input_filename)
{
#define MEMBERNAME "VectorFieldFileInput"
  infile=new_infile;
#undef MEMBERNAME
}

VectorFieldFileInput::~VectorFieldFileInput()
{
#define MEMBERNAME "~VectorFieldFileInput"
  if(infile!=NULL) {
    fclose(infile);
    infile=NULL;  // Safety
  }
#undef MEMBERNAME
}

// All children classes of VFFI should register themselves with
// RegisterChild so that they can be called as necessary from
// NewReader().
BOOL VectorFieldFileInput::RegisterChild(const VectorFieldFileInputID &id)
{
#define MEMBERNAME "RegisterChild"
  VectorFieldFileInputID *id_ptr;
  // Check for repeats
  for(id_ptr=child_list.GetFirst(); id_ptr!=NULL;
      id_ptr=child_list.GetNext()) {
    if(StringCompare(id.filetype,id_ptr->filetype)==0)
      FatalError(-1,STDDOC,"Type %s already registered",
		 (const char*)id_ptr->filetype);
  }
  child_list.Append(id);
  return 0;
#undef MEMBERNAME
}


// Clients should call NewReader() to get a VFFI instance.  The
// client is responsible for destroying (via delete) the returned
// object, i.e., treat this as a "new" operator.
VectorFieldFileInput *VectorFieldFileInput::NewReader(const char *new_file)
{
#define MEMBERNAME "NewReader"
  FILE *fptr=fopen(new_file,"rb");
  if(fptr==NULL) {
    NonFatalError(STDDOC,"Unable to open file %s for reading",new_file);
    return NULL;
  }

  // Determine file type
  VectorFieldFileInputID *id_ptr;
  for(id_ptr=child_list.GetFirst(); id_ptr!=NULL;
      id_ptr=child_list.GetNext()) {
    rewind(fptr);
    if((id_ptr->pIsType)(fptr)) break;
  }

  if(id_ptr==NULL) {
    Message("File %s is of unknown type",new_file);
    // NonFatalError(STDDOC,"File %s is of unknown type",new_file);
    fclose(fptr);
    return NULL;
  }

  rewind(fptr);

  return (id_ptr->pMakeNew)(new_file,fptr);
  // Note: It is the responsibility of the VFFI destructor to
  //       fclose fptr.
#undef MEMBERNAME
}

//////////////////////////////////////////////////////////////////////
// OmfSegmentHeader
//
const ClassDoc OmfSegmentHeader::class_doc("OmfSegmentHeader",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","24-Mar-1997");

OmfSegmentHeader::OmfSegmentHeader()
{
#define MEMBERNAME "OmfSegmentHeader"
  SetBools(0);
#undef MEMBERNAME
}

void OmfSegmentHeader::SetBools(BOOL newval)
{ // Sets all BOOL's to newval
#define MEMBERNAME "SetBools"
  title_bool=desc_bool=
    meshunit_bool=xbase_bool=ybase_bool=zbase_bool=
    xstepsize_bool=ystepsize_bool=zstepsize_bool=
    xnodes_bool=ynodes_bool=znodes_bool=
    xmin_bool=ymin_bool=zmin_bool=
    xmax_bool=ymax_bool=zmax_bool=
    valueunit_bool=valuemultiplier_bool=
    valuerangemaxmag_bool=valuerangeminmag_bool=newval;
#undef MEMBERNAME
}

BOOL OmfSegmentHeader::AllSet(void)
{ // Returns 1 iff all BOOL's are 1
#define MEMBERNAME "AllSet"
  BOOL result=title_bool*desc_bool*
    meshunit_bool*xbase_bool*ybase_bool*zbase_bool*
    xstepsize_bool*ystepsize_bool*zstepsize_bool*
    xnodes_bool*ynodes_bool*znodes_bool*
    valueunit_bool*valuemultiplier_bool*
    valuerangemaxmag_bool*valuerangeminmag_bool;

  return ( result==0 ? 0 : 1 );
#undef MEMBERNAME
}

void OmfSegmentHeader::SetFieldValue(const String &field,const String &value)
{
#define MEMBERNAME "OmfSetFieldValue"
  if(field.Length()==0) return;  // Blank line
  if(strcmp((const char *)field,"title")==0) {
    title=value;    title_bool=1;
  }
  else if(strcmp((const char *)field,"desc")==0) {
    if(desc_bool==0) { desc=value; desc_bool=1; }
    else desc.Append(value);
  }
  else if(strcmp((const char *)field,"meshunit")==0) {
    meshunit=value;    meshunit_bool=1;
  }
  else if(strcmp((const char *)field,"xbase")==0) {
    xbase=atof((const char *)value); xbase_bool=1;
  }
  else if(strcmp((const char *)field,"ybase")==0) {
    ybase=atof((const char *)value); ybase_bool=1;
  }
  else if(strcmp((const char *)field,"zbase")==0) {
    zbase=atof((const char *)value); zbase_bool=1;
  }
  else if(strcmp((const char *)field,"xstepsize")==0) {
    xstepsize=atof((const char *)value); xstepsize_bool=1;
  }
  else if(strcmp((const char *)field,"ystepsize")==0) {
    ystepsize=atof((const char *)value); ystepsize_bool=1;
  }
  else if(strcmp((const char *)field,"zstepsize")==0) {
    zstepsize=atof((const char *)value); zstepsize_bool=1;
  }
  else if(strcmp((const char *)field,"xnodes")==0) {
    xnodes=atoi((const char *)value); xnodes_bool=1;
  }
  else if(strcmp((const char *)field,"ynodes")==0) {
    ynodes=atoi((const char *)value); ynodes_bool=1;
  }
  else if(strcmp((const char *)field,"znodes")==0) {
    znodes=atoi((const char *)value); znodes_bool=1;
  }
  else if(strcmp((const char *)field,"xmin")==0) {
    xmin=atof((const char *)value); xmin_bool=1;
  }
  else if(strcmp((const char *)field,"ymin")==0) {
    ymin=atof((const char *)value); ymin_bool=1;
  }
  else if(strcmp((const char *)field,"zmin")==0) {
    zmin=atof((const char *)value); zmin_bool=1;
  }
  else if(strcmp((const char *)field,"xmax")==0) {
    xmax=atof((const char *)value); xmax_bool=1;
  }
  else if(strcmp((const char *)field,"ymax")==0) {
    ymax=atof((const char *)value); ymax_bool=1;
  }
  else if(strcmp((const char *)field,"zmax")==0) {
    zmax=atof((const char *)value); zmax_bool=1;
  }
  else if(strcmp((const char *)field,"valueunit")==0) {
    valueunit=value;   valueunit_bool=1;
  }
  else if(strcmp((const char *)field,"valuemultiplier")==0) {
    valuemultiplier=atof((const char *)value); valuemultiplier_bool=1;
  }
  else if(strcmp((const char *)field,"valuerangemaxmag")==0) {
    valuerangemaxmag=atof((const char *)value); valuerangemaxmag_bool=1;
  }
  else if(strcmp((const char *)field,"valuerangeminmag")==0) {
    valuerangeminmag=atof((const char *)value); valuerangeminmag_bool=1;
  }
  else {
    NonFatalError(STDDOC,"Unknown omf segment header field: %s (skipping)",
		  (const char *)field);
  }
#undef MEMBERNAME
}


///////////////////////////////////////////////////////////////////////////
// OmfFileInput Class

const ClassDoc OmfFileInput::class_doc("OmfFileInput",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","18-Mar-1997");

const char* OmfFileInput::file_type="OOMMF (.omf) Input File";

const VectorFieldFileInputID OmfFileInput::id(
                                OmfFileInput::file_type,
				OmfFileInput::IsType,
				OmfFileInput::MakeNew,
				VectorFieldFileInput::RegisterChild);

OmfFileInput::OmfFileInput(const char *new_filename,
			   FILE *new_infile)
  : VectorFieldFileInput(new_filename,new_infile)
{
#define MEMBERNAME "OmfFileInput"
  // Read file header here (i.e., determine # of segments)
#undef MEMBERNAME
}

VectorFieldFileInput* OmfFileInput::MakeNew(const char *new_filename,
					    FILE *new_infile)
{
  return new OmfFileInput(new_filename,new_infile);
}

BOOL OmfFileInput::IsType(FILE *testfile)
{
#define MEMBERNAME "IsType"
  char buf[SCRATCH_BUF_SIZE];
  buf[0]='\0'; // Safety
  fgets(buf,sizeof(buf),testfile);

  // Strip spaces and lowercase
  char *cptr1,*cptr2;
  cptr1=cptr2=buf;
  while(*cptr2!='\0') {
    if(!isspace(*cptr2)) *(cptr1++) = tolower(*cptr2);
    cptr2++;
  }
  *cptr1='\0';

  // Do string comparison
  if(strncmp(buf,"#oommf:rectangularmesh",
	     strlen("#oommf:rectangularmesh"))==0)
    return 1; // Correct type

  return 0;
#undef MEMBERNAME
}

INT4m OmfFileInput::ReadLine(String &buf)
{ // Reads on line from infile and sticks it into buf
  // Returns 0 on success, 1 on EOF or other error
  // NOTE: This routine is not terribly efficient (lots of
  //  unnecessary calls to string allocation/deallocation),
  //  and should only be used on omf file headers (*not* data
  //  sections), at least until the String class is written in
  //  a more efficient manner.
#define MEMBERNAME "ReadLine"
  static char readbuf[SCRATCH_BUF_SIZE];
  char *cptr;
  buf="";  // Clear buf;
  do {
    readbuf[SCRATCH_BUF_SIZE-2]='\0';  // Used for line-too-long check
    cptr=fgets(readbuf,SCRATCH_BUF_SIZE,infile);
    if(cptr==NULL) return 1; // Error
    buf.Append(readbuf,SCRATCH_BUF_SIZE);
  } while(readbuf[SCRATCH_BUF_SIZE-2]!='\0'
	&& readbuf[SCRATCH_BUF_SIZE-2]!='\n');
  return 0;
#undef MEMBERNAME
}

INT4m OmfFileInput::ParseContentLine(const String &buf,
				     String &field,String &value)
{ //   Breaks buf (import) down into field & value substrings (exports),
  // with field put into canonical form (strips whitespace & underscores,
  // and converts to lowercase).
  //   Return 0 on success, 1 on error.  The "field" value will be
  // a null string ('\0') if buf is an empty or comment line.
#define MEMBERNAME "ParseContentLine"
  field="";  value="";  // Default return values

  // Copy buf into workspace, trimming leading & trailing whitespace
  String workstr(buf);
  workstr.Trim();
  char *work=&(workstr[0]); // This is cheating!  We can't use workstr
  // again from below this point, since if we do *work may become invalid.
  // It has the advantage, though, that the memory allocated by workstr
  // is automatically freed upon function exit.

  // Tokenize workspace
  char *fieldstart=strchr(work,'#');
  if(fieldstart==NULL) return 0;  // Possibly illegal line, but assume blank
  fieldstart++;
  char *valuestart=strchr(fieldstart,':');
  if(valuestart==NULL) return 0;  // Possibly illegal line, but assume blank
  *valuestart='\0'; valuestart++;

  // Collapse field string (remove all non alpha-numeric characters) and
  // convert to lowercase.
  char *cptr1,*cptr2;
  cptr1=cptr2=fieldstart;
  while(*cptr2!='\0') {
    if(isalnum(*cptr2)) *(cptr1++) = tolower(*cptr2);
    cptr2++;
  }
  *cptr1='\0';
  field=fieldstart;

  // Skip leading whitespace in value string
  while(*valuestart!='\0' && isspace(*valuestart)) valuestart++;

  // If field is 'desc', then return as is.
  if(strncmp("desc",fieldstart,strlen("desc"))==0) {
    value=valuestart;
    return 0;
  }

  // Otherwise strip trailing comment (if any).
  cptr1=valuestart;
  while((cptr1=strchr(cptr1,'#'))!=NULL) if(*(++cptr1)=='#') break;
  if(cptr1!=NULL) *(cptr1-1)='\0';

  // Strip trailing whitespace
  cptr1=strchr(valuestart,'\0');
  while(cptr1>valuestart && isspace(*(--cptr1))) (*cptr1)='\0';

  // If field is "begin" or "end", then convert value to lowercase
  // and strip whitespace
  if(strcmp(field,"begin")==0 || strcmp(field,"end")==0) {
    cptr1=cptr2=valuestart;
    while(*cptr2!='\0') {
      if(isalnum(*cptr2)) *(cptr1++) = tolower(*cptr2);
      cptr2++;
    }
    *cptr1='\0';
  }

  value=valuestart;
  return 0;
#undef MEMBERNAME
}

// Read 3 ASCII floating point values from infile, skipping
// any leading whitespace and eating the first whitespace
// character after the third value.
// Returns 0 on success, 1 on error.
INT4m OmfFileInput::ReadTextVec3f(Vec3f &data)
{
#define MEMBERNAME "ReadTextData"
  static char buf[SCRATCH_BUF_SIZE];
  int ch,bufindex;
  int elt=0;  // Next element to fill; 0->x, 1->y, 2->z, 3->done
  REAL4 value;
  while(elt<3) {
    // Skip leading white space
    do {
      ch=fgetc(infile);
      if(ch==EOF) return 1;  // Premature end-of-file
    } while(isspace(ch));

    // Fill buffer with data up to next white space character
    bufindex=0;
    do {
      if(ch=='d' || ch=='D') ch='e';  // Convert FORTRAN 'd' exponent
      buf[bufindex++]=ch;             // designator to C 'e' designator
      ch=fgetc(infile);
      if(ch==EOF) return 1;  // Premature end-of-file
    } while(!isspace(ch) && bufindex<SCRATCH_BUF_SIZE-1);
    if(bufindex>=SCRATCH_BUF_SIZE-1) {
      // Single data elt too big to fit in buffer?!
      NonFatalError(STDDOC,"Bad data in file %s;"
		    " Single value >%d characters long",
		    (const char *)filename,SCRATCH_BUF_SIZE);
      return 1;
    }
    buf[bufindex]='\0';

    // Convert to floating point and store
    value=atof(buf);
    if(elt==0)       data.x=value;
    else if(elt==1)  data.y=value;
    else             data.z=value;
    elt++;
  }

  return 0;
#undef MEMBERNAME
}

Mesh *OmfFileInput::NewMesh()
{
#define MEMBERNAME "NewMesh"
  String buf,field,value;
  // Skip preamble
  while(1) {
    if(ReadLine(buf)!=0) {
      NonFatalError(STDDOC,"Premature EOF in file %s "
		    "(no segment header found)",
		    (const char *)filename);
      return new EmptyMesh();
    }
    ParseContentLine(buf,field,value);
    if(strcmp(field,"begin")==0  && strcmp(value,"header")==0) break;
  }

  // Read head and store results into OmfSegmentHeader struct
  OmfSegmentHeader omfseghead;
  omfseghead.SetBools(0);  // Safety
  while(1) {
    if(ReadLine(buf)!=0) {
      NonFatalError(STDDOC,"Premature EOF in file %s "
		    "(missing segment header end mark)",
		    (const char *)filename);
      return new EmptyMesh();
    }
    ParseContentLine(buf,field,value);
    if(strcmp(field,"end")==0 && strcmp(value,"header")==0) break;
    if(field.Length()>0) omfseghead.SetFieldValue(field,value);
  }
  if(!omfseghead.AllSet()) {
      NonFatalError(STDDOC,"Incomplete segment header in file %s; Skipping",
		    (const char *)filename);
      return new EmptyMesh();
  }

  // Skip ahead in file to data section
  while(1) {
    if(ReadLine(buf)!=0) {
      NonFatalError(STDDOC,"Premature EOF in file %s (no data section found)",
		    (const char *)filename);
      return new EmptyMesh();
    }
    ParseContentLine(buf,field,value);
    if(strcmp(field,"begin")==0  && strcmp(value,"datatext")==0) break;
  }

  // Set up auxiliary data
  BoundingBox range;
  Vec3f corner1(omfseghead.xmin,omfseghead.ymin,omfseghead.zmin);
  Vec3f corner2(omfseghead.xmax,omfseghead.ymax,omfseghead.zmax);
  range.Set(corner1,corner2);
  List<Vec3f> boundary;

  REAL8 zave=(omfseghead.zmin+omfseghead.zmax)/2.0;
  boundary.SetFirst(Vec3f(omfseghead.xmin,omfseghead.ymin,zave));
  boundary.Append(  Vec3f(omfseghead.xmin,omfseghead.ymax,zave));
  boundary.Append(  Vec3f(omfseghead.xmax,omfseghead.ymax,zave));
  boundary.Append(  Vec3f(omfseghead.xmax,omfseghead.ymin,zave));
  boundary.Append(  Vec3f(omfseghead.xmin,omfseghead.ymin,zave));

  // Initialize mesh
  GridVec3f *mesh=new GridVec3f((const char *)filename,omfseghead.xnodes,
				omfseghead.ynodes,omfseghead.znodes,
				range,boundary);

  // Set data normalization value
  double norm=omfseghead.valuerangemaxmag;
  if(norm>SQRT_EPSILON) norm=1./norm;
  else norm=1.0;

  // Read data
  INT4m i,j,k;
  Vec3f data;
  for(k=0;k<omfseghead.znodes;k++)
    for(j=0;j<omfseghead.ynodes;j++)
      for(i=0;i<omfseghead.xnodes;i++) {
	// Read one entry
	if(ReadTextVec3f(data)!=0) {
	  NonFatalError(STDDOC,"Premature EOF in file %s in data section",
			(const char *)filename);
	  delete mesh;
	  return new EmptyMesh();
	}
	data*=norm; // Normalize data.
	(*mesh)(i,j,k)=data;
      }

  // Skip to segment end
  while(1) {
    if(ReadLine(buf)!=0) {
      NonFatalError(STDDOC,"Premature EOF in file %s (no segment end marker)",
		    (const char *)filename);
      return mesh;
    }
    ParseContentLine(buf,field,value);
    if(strcmp(field,"end")==0  && strcmp(value,"segment")==0) break;
  }

  // NOTE: *infile is closed by base destructor
  // (Moreover, there may be additional segments in this file.)

  return mesh;
#undef MEMBERNAME
}


///////////////////////////////////////////////////////////////////////////
// VioFileInput Class

const ClassDoc VioFileInput::class_doc("VioFileInput",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","19-Mar-1997");

const char* VioFileInput::file_type="VIO (.vio) Input File";

const VectorFieldFileInputID VioFileInput::id(
                                VioFileInput::file_type,
				VioFileInput::IsType,
				VioFileInput::MakeNew,
				VectorFieldFileInput::RegisterChild);

VioFileInput::VioFileInput(const char *new_filename,
			   FILE *new_infile)
  : VectorFieldFileInput(new_filename,new_infile)
{
#define MEMBERNAME "VioFileInput"
  // Read header here
#undef MEMBERNAME
}

VectorFieldFileInput* VioFileInput::MakeNew(const char *new_filename,
					    FILE *new_infile)
{
  return new VioFileInput(new_filename,new_infile);
}

BOOL VioFileInput::IsType(FILE *testfile)
{
#define MEMBERNAME "IsType"
  const char   VecFileMagic02[8] = "VecFil\x02";
  char buf[8]="\0\0\0\0\0\0\0"; // Safety init
  fread(buf,1,8,testfile);
  if(memcmp(VecFileMagic02,buf,8)==0) return 1;  // Match!
  return 0;
#undef MEMBERNAME
}

Mesh *VioFileInput::NewMesh()
{
#define MEMBERNAME "NewMesh"
  Mesh *newmesh=NewRectGrid();
  if(newmesh==NULL) newmesh=new EmptyMesh();
  return newmesh;
#undef MEMBERNAME
}

GridVec3f *VioFileInput::NewRectGrid()
{
#define MEMBERNAME "NewRectGrid"
  VioFile viofile;

  // Open file and get grid dimensions
  if(viofile.SetFilePtr(infile)!=0) return NULL;
  INT4m xsize,ysize,zsize;
  viofile.GetDimensions(xsize,ysize,zsize);

  // Set bounding box and boundary list
  // Note: Vio files sample at cell centers
  BoundingBox range;
  List<Vec3f> boundary;
  range.Set(Vec3f(-0.5,-0.5,-0.5),
  	  Vec3f(REAL4(xsize-0.5),REAL4(ysize-0.5),REAL4(zsize-0.5)));
  boundary.SetFirst(Vec3f(-0.5,-0.5,0.0));
  boundary.Append(Vec3f(REAL4(xsize-0.5),-0.5,0.0));
  boundary.Append(Vec3f(REAL4(xsize-0.5),REAL4(ysize-0.5),0.0));
  boundary.Append(Vec3f(-0.5,REAL4(ysize-0.5),0.0));
  boundary.Append(Vec3f(-0.5,-0.5,0.0));

  // Initialize mesh
  GridVec3f *mesh=new GridVec3f(filename,xsize,ysize,zsize,range,boundary);

  // Fill grid
  viofile.FillArray(*mesh);

  // NOTE: *infile is closed by base destructor

  return mesh;
#undef MEMBERNAME
}
