/* FILE: vecfile.h                 -*-Mode: c++-*-
 *
 * Header file for vector field file I/O.
 *
 */

#include <stdio.h>
#include "machine.h"
#include "misc.h"
#include "template.h"
#include "vector.h"

class VectorFieldFileInput;  // Forward declaration

//////////////////////////////////////////////////////////////////////
// VectorFieldFileInputID
//
class VectorFieldFileInputID
{
  friend VectorFieldFileInput;
private:
  static const ClassDoc class_doc;
  String filetype;
  BOOL (*pIsType)(FILE *input_file);
  VectorFieldFileInput* (*pMakeNew)(const char *new_filename,
				    FILE *new_infile);
  void Dup(const VectorFieldFileInputID &rhs); // Copy
public:
  // Constructor *with* registration (for global initialization)
  // (Set last field to NULL to avoid registration.)
  VectorFieldFileInputID(const char *new_filetype,
			 BOOL (*newpIsType)(FILE *),
			 VectorFieldFileInput *
			    (*newpNewReader)(const char *,FILE *),
			 BOOL (*pRegister)(const VectorFieldFileInputID &)
			 );

  // Copy constructor
  VectorFieldFileInputID(const VectorFieldFileInputID &copyobj);

  VectorFieldFileInputID(); // Default constructor; Needed by List<T>
  VectorFieldFileInputID& operator=(const VectorFieldFileInputID &rhs);

  ~VectorFieldFileInputID();
};

//////////////////////////////////////////////////////////////////////
// VectorFieldFileInput
//
class VectorFieldFileInput
{
  // NOTES ON CHILD CLASSES:
  //
  // Every child class should have a static member function defined
  // like this:
  //
  // static VectorFieldFileInput* MakeNew(const char *new_filename,
  //                                      FILE *new_infile)
  //
  // which calls its "internal" constructor.  This function
  // is passed to VectorFieldFileInput::RegisterChild as part
  // of a VectorFieldFileInputID.
  //
  // All children classes should have a static member variable:
  //
  //   static VectorFieldFileInputID id;
  //
  // This initialization of this variable should call
  // VectorFieldFileInput::RegisterChild(id) so that they can be
  // called as necessary from VectorFieldFileInput::NewReader().
  //
  // The FILE *infile is closed in the base class destructor

private:
  static const ClassDoc class_doc;
  static List<VectorFieldFileInputID> child_list;
  // The following 3 functions are left undefined on purpose
  VectorFieldFileInput(); // Default constructor
  VectorFieldFileInput(VectorFieldFileInput &copyobj);  // Copy constructor
  VectorFieldFileInput& operator=(VectorFieldFileInput &rhs); // Assignment op

protected:
  String filename;
  FILE *infile;

  // Internal constructor
  VectorFieldFileInput(const char *new_filename,FILE *new_infile);

  static BOOL RegisterChild(const VectorFieldFileInputID &id);

public:
  // Clients should call NewReader() to get a VFFI instance.  The
  // client is responsible for destroying (via delete) the returned
  // object, i.e., treat this as a "new" operator.
  static VectorFieldFileInput *NewReader(const char *new_filename);

  virtual ~VectorFieldFileInput();
  const char *GetFileName(void) { return (const char *)filename; }
  virtual const char *FileType(void)=0;
  virtual Mesh *NewMesh(void)=0;  // Derived classes should return
    // a *EmptyMesh on error (not NULL).  Either way, the client
    // the client is responsible for destroying (via delete) the
    // returned mesh.
};

//////////////////////////////////////////////////////////////////////
// OmfSegmentHeader
//
struct OmfSegmentHeader
{
private:
  static const ClassDoc class_doc;
public:
  // HEADER VALUES
  String title; // Long file name or title goes here
  String desc;  // File description
  String meshunit; // Fundamental mesh measurement unit.  Treated as a label.
                   // Example: nm

  REAL8 xbase,ybase,zbase; // (xbase,ybase,zbase) is the position, in
                           // 'meshunit', of the first point in the data
                           //  section (below).

  REAL8 xstepsize,ystepsize,zstepsize;  // Distance between adjacent
  //grid pts.: on the x-axis, 20 nm, etc.  The sign on these values
  // determines the grid orientation relative to (xbase,ybase,zbase).

  INT4 xnodes,ynodes,znodes; // Number of nodes along each axis.

  REAL8 xmin,ymin,zmin;  // Corner points defining mesh bounding box
  REAL8 xmax,ymax,zmax;  // and boundary.

  String valueunit; // Fundamental field value unit, treated as a label
                    // Example: kA/m

  REAL8 valuemultiplier; // Multiply file values by this to get
                         // true value in 'valueunits'.

  REAL8 valuerangemaxmag,valuerangeminmag; // These are in file value
  // units, and are used as hints (or defaults) by postprocessing programs.
  //  The mmdisp program ignores any points with magnitude smaller than
  // valuerangeminmag, and uses valuerangemaxmag to scale inputs for display.
  // The use of these values is postprocessor dependent, and should be
  // interactively overrideable, but typically ValueRangeMaxMag should be
  // the maximum vector magnitude in the file data, and ValueRangeMinMag
  // should be a lower bound on non-zero file vector magnitudes.

  // BOOLEANS
  // The following boolean values indicate whether each of the above values
  // have been set (1=>set, 0=>not set).
  BOOL title_bool,desc_bool;
  BOOL meshunit_bool,xbase_bool,ybase_bool,zbase_bool;
  BOOL xstepsize_bool,ystepsize_bool,zstepsize_bool;
  BOOL xnodes_bool,ynodes_bool,znodes_bool;
  BOOL xmin_bool,ymin_bool,zmin_bool;
  BOOL xmax_bool,ymax_bool,zmax_bool;
  BOOL valueunit_bool,valuemultiplier_bool;
  BOOL valuerangemaxmag_bool,valuerangeminmag_bool;

  // Utility routines
  OmfSegmentHeader();
  void SetBools(BOOL newval);  // Sets all BOOL's to newval
  BOOL AllSet(); // Returns 1 iff all BOOL's are 1
  void SetFieldValue(const String &field,const String &value);
};

//////////////////////////////////////////////////////////////////////
// OmfFileInput
//
class OmfFileInput : public VectorFieldFileInput
{
private:
  static const ClassDoc class_doc;
  static const char *file_type;
  static const VectorFieldFileInputID id;

  // File read helper routines
  INT4m ReadLine(String &buf);
  INT4m ParseContentLine(const String &buf,String &field,String &value);
  INT4m ReadTextVec3f(Vec3f &data);

  // Internal constructor
  OmfFileInput(const char *new_filename,FILE *new_infile);

  // The following 3 functions are left undefined on purpose
  OmfFileInput(OmfFileInput &copyobj);  // Copy constructor
  OmfFileInput& operator=(OmfFileInput &rhs); // Assignment op

public:
  OmfFileInput(); // Default constructor.  DON'T USE THIS!
  // This is a non-existent constructor that I would prefer to
  // keep in the 'private' section, except that some C++ compilers
  // (I won't name names, but their initials are S, G, and I) require
  // at least 1 public constructor.  Sigh...
  static VectorFieldFileInput* MakeNew(const char *new_filename,
				       FILE *new_infile);
  const char *FileType(void) { return file_type; }
  static BOOL IsType(FILE *input_file);
  Mesh *NewMesh(void);
};

//////////////////////////////////////////////////////////////////////
// VioFileInput
//
class VioFileInput : public VectorFieldFileInput
{
private:
  static const ClassDoc class_doc;
  static const char *file_type;
  static const VectorFieldFileInputID id;

  // The following 3 functions are left undefined on purpose
  VioFileInput(); // Default constructor
  VioFileInput(OmfFileInput &copyobj);  // Copy constructor
  VioFileInput& operator=(OmfFileInput &rhs); // Assignment op

public:
  // Internal constructor
  VioFileInput(const char *new_filename,FILE *new_infile);

  // Standard constructor
  static VectorFieldFileInput* MakeNew(const char *new_filename,
				       FILE *new_infile);
  const char *FileType(void) { return file_type; }
  static BOOL IsType(FILE *input_file);
  Mesh *NewMesh(void); // Returns *EmptyMesh on error
  GridVec3f *NewRectGrid(void);  // Return NULL on error
};
