/* FILE: vector.cc                -*-Mode: c++-*-
 *
 * Vector and collections of vectors
 *
 */

#include <stdio.h>

#include "vector.h"
#include "fileio.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

//////////////////////////////////////////////////////////////////////////
// 4-byte floating point vectors
const ClassDoc Vec3f::class_doc("Vec3f",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","11-Dec-1996");

//////////////////////////////////////////////////////////////////////////
// 3D rectangular solid
const ClassDoc BoundingBox::class_doc("BoundingBox",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","13-Dec-1996");

BOOL BoundingBox::IsIn(const Vec3f &location) const
{
#define MEMBERNAME "IsIn"
  return (location.Bigger(min) && (max.Bigger(location)));
#undef MEMBERNAME
}


//////////////////////////////////////////////////////////////////////////
// Regular 3D grid of Vec3f's
const ClassDoc GridVec3f::class_doc("GridVec3f",
		    "Michael J. Donahue (michael.donahue@nist.gov)",
		    "1.0.0","11-Dec-1996");

// The following constructor should be called from file input functions.
GridVec3f::GridVec3f(const char *filename,
		     INT4m newi,INT4m newj,INT4m newk,
		     BoundingBox &newrange,List<Vec3f> &newboundary)
{
  Filename=new char[strlen(filename)+1];
  strcpy(Filename,filename);
  SetSize(newi,newj,newk);
  range=newrange;
  boundary=newboundary;
}

INT4m GridVec3f::GetDisplayList(REAL4m &xstep_request,
				REAL4m &ystep_request,
				REAL4m &zstep_request,
				const BoundingBox &dsprange,
				List<LocatedVector> &display_list) const
{ // Fills display_list with points from mesh inside "dsprange", with
  // step size close to ?step_request.
  // Use ?step_request=0 to remove constraint on that axis, e.g.,
  // ?step_request is reset to value actually used.
  // if xstep_request=ystep_request=zstep_request=0, then this
  // routine places all points inside "dsprange" into display_list.
  // Currently, this routine adjusts ?step_request to integral values,
  // because non-integral values lead to unattractive banding in the
  // output display list.
#define MEMBERNAME "GetDisplayList"
  if(xstep_request<0 || ystep_request<0 || zstep_request<0)
    FatalError(-1,STDDOC,"%s: xstep=%f, ystep=%f, zstep=%d",
	       ErrBadParam,xstep_request,ystep_request,zstep_request);

  display_list.Clear();
  INT4m isize,jsize,ksize;
  GetSize(isize,jsize,ksize);
  if(isize<1 || jsize<1 || ksize<1)
    return display_list.GetSize(); // Empty grid
  Vec3f rangemin,rangemax;
  dsprange.GetExtremes(rangemin,rangemax);

  INT4m ifirst,ilast,jfirst,jlast,kfirst,klast;
  ifirst=INT4m(ceil(rangemin.x)); ilast=INT4m(floor(rangemax.x))+1;
  jfirst=INT4m(ceil(rangemin.y)); jlast=INT4m(floor(rangemax.y))+1;
  kfirst=INT4m(ceil(rangemin.z)); klast=INT4m(floor(rangemax.z))+1;

  if(ifirst<0 || ilast>isize || jfirst<0 || jlast>jsize ||
     kfirst<0 || klast>ksize) {
    NonFatalError(STDDOC,"Range box out-of-bounds");
    if(ifirst<0) ifirst=0; if(ilast>isize) ilast=isize;
    if(jfirst<0) jfirst=0; if(jlast>jsize) jlast=jsize;
    if(kfirst<0) kfirst=0; if(klast>ksize) klast=ksize;
  }

  // Calculate step sizes
  // Note: Non-integer step size produce ugly banding in output sampling
  xstep_request=rint(xstep_request);
  ystep_request=rint(ystep_request);
  zstep_request=rint(zstep_request);

  REAL4m max_step=MAX(MAX(isize,jsize),ksize);
  if(xstep_request>max_step) xstep_request=max_step;
  if(ystep_request>max_step) ystep_request=max_step;
  if(zstep_request>max_step) zstep_request=max_step;

  INT4m istep=INT4m(xstep_request);
  INT4m jstep=INT4m(ystep_request);
  INT4m kstep=INT4m(zstep_request);

  if(istep<1) istep=1;  if(jstep<1) jstep=1;  if(kstep<1) kstep=1;

  // Determine offsets
  INT4m ioff=((ilast-ifirst-1)%istep)/2;
  INT4m joff=((jlast-jfirst-1)%jstep)/2;
  INT4m koff=((klast-kfirst-1)%kstep)/2;

  INT4m i,j,k;
  Vec3f position;
  for(k=kfirst+koff;k<klast;k+=kstep) {
    position.z=REAL4m(k);
    for(j=jfirst+joff;j<jlast;j+=jstep) {
      position.y=REAL4m(j);
      for(i=ifirst+ioff;i<ilast;i+=istep) {
	position.x=REAL4m(i);
	display_list.Append(LocatedVector(position,GridVec(i,j,k)));
      }
    }
  }

  return display_list.GetSize();
#undef MEMBERNAME
}
