/* FILE: vector.h                 -*-Mode: c++-*-
 *
 * Vector and collections of vectors
 *
 */

#ifndef VECTOR_H
#define VECTOR_H

#include "misc.h"
#include "template.h"

/* End includes */     /* Optional directive to build.tcl */

//////////////////////////////////////////////////////////////////////////
// 4-byte floating point vectors
class Vec3f
{
private:
  static const ClassDoc class_doc;
public:
  REAL4 x,y,z;  // Components
  void Set(REAL4 nx,REAL4 ny,REAL4 nz) { x=nx; y=ny; z=nz; }
  Vec3f(void) { x=y=z=0.0; } // Default constructor
  Vec3f(REAL4 nx,REAL4 ny,REAL4 nz) { Set(nx,ny,nz); }
  Vec3f(const Vec3f &v) : x(v.x), y(v.y), z(v.z) {} // Copy constructor
  // Assignment operator
  Vec3f& operator=(const Vec3f &v) { x=v.x; y=v.y; z=v.z; return *this; }
  BOOL Equals(const Vec3f &v) const { return (x==v.x  && y==v.y && z==v.z); }
  BOOL Bigger(const Vec3f &v) const { return (x>=v.x  && y>=v.y && z>=v.z); }
  inline REAL4 MagSq() const { return x*x+y*y+z*z; }
  inline REAL4 Mag() const { return sqrt(MagSq()); }
  inline Vec3f & operator+=(const Vec3f& v);
  inline Vec3f & operator-=(const Vec3f& v);
  inline Vec3f & operator*=(REAL4 a);
  friend Vec3f operator+(const Vec3f& v,const Vec3f& w);
  friend Vec3f operator-(const Vec3f& v,const Vec3f& w);
  friend Vec3f operator*(REAL4 a,const Vec3f& v);
  friend Vec3f operator*(const Vec3f& v,REAL4 a);
};

Vec3f & Vec3f::operator+=(const Vec3f& v) {
  x+=v.x; y+=v.y; z+=v.z; return *this;
}
Vec3f & Vec3f::operator-=(const Vec3f& v) {
  x-=v.x; y-=v.y; z-=v.z; return *this;
}
Vec3f & Vec3f::operator*=(REAL4 a)  { x*=a; y*=a; z*=a; return *this; }

inline Vec3f operator+(const Vec3f& v,const Vec3f& w)
{
  Vec3f temp;
  temp.x=v.x+w.x;  temp.y=v.y+w.y;  temp.z=v.z+w.z;
  return temp;
}

inline Vec3f operator-(const Vec3f& v,const Vec3f& w)
{
  Vec3f temp;
  temp.x=v.x-w.x;  temp.y=v.y-w.y;  temp.z=v.z-w.z;
  return temp;
}

inline Vec3f operator*(REAL4 a,const Vec3f& v)
{
  Vec3f temp;
  temp.x=a*v.x;  temp.y=a*v.y;  temp.z=a*v.z;
  return temp;
}

inline Vec3f operator*(const Vec3f& v,REAL4 a) { return a*v; }


//////////////////////////////////////////////////////////////////////////
// 4-byte floating point Position _and_ Vector
struct LocatedVector
{
public:
  Vec3f location;
  Vec3f value;
  LocatedVector() { location.Set(0.,0.,0.); value.Set(0.,0.,0.); }
  LocatedVector(const Vec3f &newloc,const Vec3f &newval) {
    location=newloc; value=newval;
  }
};

//////////////////////////////////////////////////////////////////////////
// Bounding Box
class BoundingBox
{
  static const ClassDoc class_doc;
  Vec3f min;
  Vec3f max;
public:
  void SortAndSet(const Vec3f &corner1,const Vec3f &corner2) {
    if(corner1.x<corner2.x) { min.x=corner1.x; max.x=corner2.x; }
    else                    { min.x=corner2.x; max.x=corner1.x; }
    if(corner1.y<corner2.y) { min.y=corner1.y; max.y=corner2.y; }
    else                    { min.y=corner2.y; max.y=corner1.y; }
    if(corner1.z<corner2.z) { min.z=corner1.z; max.z=corner2.z; }
    else                    { min.z=corner2.z; max.z=corner1.z; }
  }
  void Set(const Vec3f &corner1,const Vec3f &corner2) {
    min.x=corner1.x; max.x=corner2.x;
    min.y=corner1.y; max.y=corner2.y;
    min.z=corner1.z; max.z=corner2.z;
  }
  // Default constructor
  BoundingBox(void) { min.Set(0.,0.,0.); max.Set(0.,0.,0.); }
  // Copy constructor
  BoundingBox(const BoundingBox& newbb) : min(newbb.min), max(newbb.max) {}
  // Assignment operator
  BoundingBox& operator=(const BoundingBox& bb) { 
    min=bb.min; max=bb.max; return *this;
  }
  BOOL IsIn(const Vec3f &location) const;
  void GetMinPt(Vec3f &minpt) const { minpt=min; }
  void GetExtremes(Vec3f &minpt,Vec3f &maxpt) const { minpt=min; maxpt=max; }
  REAL4m GetWidth(void)  const { return max.x-min.x; }
  REAL4m GetHeight(void) const { return max.y-min.y; }
  REAL4m GetDepth(void)  const { return max.z-min.z; }
  REAL4m GetVolume(void) const { return GetWidth()*GetHeight()*GetDepth(); }
  Vec3f GetDiag(void) { return max-min; }
};


//////////////////////////////////////////////////////////////////////////
// Mesh (abstract) class
class Mesh
{
public:
  virtual const char* GetName() const { return ""; }
  virtual INT4m GetRange(BoundingBox &range) const =0;
  virtual INT4m GetBoundaryList(List<Vec3f> &boundary_list) const =0;
  virtual REAL4m GetApproximateCellSize(void) const =0;
  virtual REAL4m GetPreferredStepSize(REAL4m step) const { return rint(step); }
  virtual REAL4m GetBaseIncrement(void) const { return 1.; }
  virtual INT4m GetDisplayList(REAL4m &xstep_request,
			       REAL4m &ystep_request,
			       REAL4m &zstep_request,
			       const BoundingBox &range,
			       List<LocatedVector> &display_list) const =0;
  virtual ~Mesh() {}
};

//////////////////////////////////////////////////////////////////////////
// Empty mesh
class EmptyMesh:public Mesh
{
private:
  BoundingBox empty_range;
  List<Vec3f> empty_boundary_list;
public:
  EmptyMesh() {}
  const char* GetName() const { return "<empty>"; }
  INT4m GetRange(BoundingBox &range) const { range=empty_range; return 0; }
  INT4m GetBoundaryList(List<Vec3f> &boundary_list) const {
    boundary_list=empty_boundary_list;
    return 0;
  }
  REAL4m GetApproximateCellSize(void) const { return 1; }
  INT4m GetDisplayList(REAL4m &,REAL4m &,REAL4m &,
		       const BoundingBox&,
		       List<LocatedVector> &display_list) const {
    display_list.Clear();
    return 0;
  }
};

//////////////////////////////////////////////////////////////////////////
// Regular 3D grid of Vec3f's
class GridVec3f:public Mesh
{ // NOTE: dim1,dim2,dim3 of grid should be thought of in the order z,y,x
  // (or k,j,i), because in applications typically z-dim (k) will be smallest,
  // x-dim (i) will be largest, so the data is stored arr[k][j][i].  This is
  // why we overload the () operator to map (i,j,k) -> [k][j][i].
  // IN PARTICULAR, efficient loops will increment the innermost loop on i.
  //   SO, KEEP THIS STRAIGHT: (i,j,k) coming in to member functions
  // should access grid[k][j][i], or (better) use GridVec(i,j,k).
private:
  static const ClassDoc class_doc;
  char *Filename;
  Array3D<Vec3f> grid;     // Magnetization vector array
  inline INT4m SetSize(INT4m newi,INT4m newj,INT4m newk) {
    return grid.Allocate(newk,newj,newi);
  }
  BoundingBox range;
  List<Vec3f> boundary;

  // The following function is left undefined on purpose
  GridVec3f& operator=(GridVec3f& rhs); // Assignment op

public:
  inline INT4m GetSize(INT4m &isize,INT4m &jsize,INT4m &ksize) const {
    return grid.GetSize(ksize,jsize,isize);
  }
  inline INT4m GetSize() const { return grid.GetSize();  }
  GridVec3f(void) { Filename=NULL; }

  // The following constructor should be called from file input functions.
  GridVec3f(const char *filename,INT4m newi,INT4m newj,INT4m newk,
	    BoundingBox &newrange,List<Vec3f> &newboundary);

  GridVec3f(GridVec3f &other) { // Copy constructor
    if(other.Filename==NULL) Filename=NULL;
    else {
      Filename=new char[strlen(other.Filename)+1];
      strcpy(Filename,other.Filename);
    }
    grid.Copy(other.grid);
  }
  virtual inline const char *GetName() const { return Filename; }
  inline Vec3f& GridVec(INT4m i,INT4m j,INT4m k) { return grid(k,j,i); }
  inline Vec3f& operator()(INT4m i,INT4m j,INT4m k) { return GridVec(i,j,k); }
  inline const Vec3f& GridVec(INT4m i,INT4m j,INT4m k) const {
    return grid(k,j,i);
  }
  inline const Vec3f& operator()(INT4m i,INT4m j,INT4m k) const {
    return GridVec(i,j,k);
  }
  virtual INT4m GetRange(BoundingBox &myrange) const {
    myrange=range; return 0;
  }
  virtual INT4m GetBoundaryList(List<Vec3f> &boundary_list) const {
    boundary_list=boundary;
    return 0;
  }
  virtual REAL4m GetApproximateCellSize(void) const { return 1.0; }
  virtual INT4m GetDisplayList(REAL4m &xstep_request,
			       REAL4m &ystep_request,
			       REAL4m &zstep_request,
			       const BoundingBox &dsprange,
			       List<LocatedVector> &display_list) const;
  virtual ~GridVec3f(void) { if(Filename!=NULL) delete[] Filename; }
};

#endif  // VECTOR_H


