#!/bin/sh
# FILE: build.tcl
######################################################################
# INTRODUCTION
######################################################################
# This is a tcl script which controls the compilation, linking, and
# installation of executable software and documentation from sources.
# Its function is similar to the Unix utility "make".

######################################################################
# CONFIGURATION
######################################################################
# In order to work properly on a variety of computing platforms, this
# script needs the values of several Tcl variables to be set appropriately 
# as a description of your computing environment.  That configuration task
# is assigned to the scripts in the ./build subdirectory.  Please see the
# file ./build/config.tcl for configuration instructions.

######################################################################
# BOOTSTRAP
######################################################################
# The following snippet is the Tcl bootstrap trick on Unix.
#
# This comment continues on in Tcl, but not in sh \
exec tclsh "$0" ${1+"$@"}
#
# If the above exec doesn't work, replace ${1+"$@"} with "$@"
# The syntax ${1+"$@"} means: if $1 (positional parameter #1)
# is set, then substitute "$@" (all parameters as seperate words),
# if $1 is not set (no command line arguments), then substitute
# nothing.  Bash makes no substitution for "$@" if $1 is not set,
# so these are equivalent under bash.  Apparently there is some
# danger that a shell may substitute "" for "$@" if $1 is not set,
# but I don't know which shells do that.

######################################################################
# LIBRARY OF TCL SUPPORT PROCS
######################################################################
# This script depends on library routines stored in the ./build
# subdirectory.  Tell the auto-loader where to find them.
lappend auto_path [file dirname [info script]]/build

######################################################################
# BUILD DEFINITIONS
######################################################################
# Defining these lists defines the build task to be performed in this
# directory.

# List the subdirectories in which software builds must be performed 
# before a build in this directory.  If a particular sequence is required
# for building in the subdirectories, list them in that sequence.
set PRESUBDIRS {}

# List the subdirectories in which software builds must be performed 
# after a build in this directory.  If a particular sequence is required
# for building in the subdirectories, list them in that sequence.
set POSTSUBDIRS {}

# List the file name bases of any executables (targets) to be built 
# from source code in this directory.
set TARGETS {}

# List the file name bases of any C++ source code files in this directory
# which contribute to the building of one of the TARGETS.
set SOURCES {oommfInit}

# Each *.cc source code is run through proc FindIncludeDep to determine
# non-system #include level dependencies.  You can include any additional
# dependencies in EXTINCL
set EXTINCL {}

# The name of the directory (relative to this directory) in which to
# build the object modules and executables
set BINDIR bin

# A list of directories in addition to the directory containing this
# script in which user include files (those which appear in the 
# form ``#include "foo.h"'' in the source code) may be found.
set INCLUDEDIRS {}

# A list of paths relative to this directory of object files which
# must be linked against to build the TARGETS, but which are not
# generated by compiling SOURCES.
# 
# Try to come up with a better solution than this... 
set OTHEROBJECTS {}

######################################################################
# DEFAULT CONFIGURATION VALUES
######################################################################
# Collect configuration information about the computing environment
# from another script, ./build/defatuls.tcl .
# Ought to use file join here
EvaluateScript [file dirname [info script]]/build/defaults.tcl

######################################################################
# Place here any modifications to the configuration globals required
# for this build.

# Check for "debug" request
if {[string compare $argv "debug"]==0} {
    set DEBUGLVL 100
    AppendUnique CFLAGS $DEBUGFLAG
    AppendUnique CFLAGS $STRICTWARNFLAG
} else {
    set DEBUGLVL 0
}

# Import the debug level macro definition into the compiler
AppendUnique CFLAGS ${MACRODEFINEFLAG}DEBUGLVL=$DEBUGLVL

# Visual C++ needs our help defining a MACHINE_TYPE macro for
# machine.h
if {[string compare $PLATFORM windows]==0} {
  AppendUnique CFLAGS ${MACRODEFINEFLAG}MACHINE_TYPE=5
}

######################################################################
# Actually do the build
DoBuild
