/* FILE: misc.h                 -*-Mode: c++-*-
 *
 * Useful miscellaneous stuff.
 *
 */

#ifndef MISC_H
#define MISC_H

#include "machine.h"

#include <math.h>
#include <stdarg.h>

/* End includes */     /* Optional directive to build.tcl */

#define EPSILON        2.22045e-16
#define SQRT_EPSILON   1.49012e-08
#define BIG_VALUE 8.988465674311578540726371186585e+307

#ifndef PI
#define PI             3.14159265358979323846
#endif

#ifndef SQRT2
#define SQRT2          1.41421356237309504880
#endif

#define SCRATCH_BUF_SIZE 1024  // Size of scratch space buffers

extern int Verbosity;
#ifndef DEBUGLVL
#define DEBUGLVL 100  // 0 => No debug code, 100 => Full debug code
#endif 

// Several useful macros
#define SQ(x) ((x)*(x))
#define MIN(x,y) ((x)<(y)?(x):(y))
#define MAX(x,y) ((x)>(y)?(x):(y))

//////////////////////////////////////////////////////////////////////////
// Class documentation structure
struct ClassDoc {
public:
  const CHAR *classname;
  const CHAR *maintainer;
  const CHAR *revision;
  const CHAR *revdate;
  ClassDoc(const CHAR *new_classname,const CHAR *new_maintainer,
	   const CHAR *new_revision,const CHAR *new_revdate);
};

//////////////////////////////////////////////////////////////////////////
// Error codes
extern const char* ErrArrOutOfBounds;
extern const char* ErrBadFile;
extern const char* ErrBadFileHeader;
extern const char* ErrBadFileOpen;
extern const char* ErrInputEOF;
extern const char* ErrOutputEOF;
extern const char* ErrBadParam;
extern const char* ErrDispInit;
extern const char* ErrMisc;
extern const char* ErrNoMem;
extern const char* ErrOverflow;
extern const char* ErrProgramming;

//////////////////////////////////////////////////////////////////////////
// Error handlers
// Note: Use preprocessor macros __FILE__ and __LINE__ to get
//       'filename' and 'lineno' respectively.
void FatalError(INT4m errorcode,
		const CHAR *membername,
		const ClassDoc &class_doc,
		const CHAR *filename,INT4m lineno,
		const CHAR *fmt,...);
// Doesn't return; exits with code 'errorcode'

void NonFatalError(const CHAR *membername,
		   const ClassDoc &class_doc,
		   const CHAR *filename,INT4m lineno,
		   const CHAR *fmt,...);
// Prints warning message and returns.

// Versions of the above for non-class functions
void PlainWarning(const CHAR *fmt,...);
void PlainError(INT4m errorcode,const CHAR *fmt,...);

// Routine to print message to standard (user) output channel
int Message(const CHAR *message_fmt,...);

// C-interface to Tcl/Tk error boxes.  The above routines map
// into this.  This routine is defined in tkcmds.cc
enum MessageType { MSG_INFO, MSG_WARNING, MSG_ERROR };
INT4m TkMessage(MessageType mt,const CHAR *msg,INT4m exitcode=1);

// If you #define MEMBERNAME at the start of each member definition,
// then you can use the following macro to expand to 4 of the fields
// needed by the above error handlers.
#define STDDOC MEMBERNAME,class_doc,__FILE__,__LINE__

//////////////////////////////////////////////////////////////////////////
// Byte reversal routines, for conversion between big-endian
// and little-endian representations
inline UINT2 flip(UINT2 input);
inline INT2  flip(INT2  input);
inline INT4  flip(INT4  input);
inline REAL4 flip(REAL4 input);
inline REAL8 flip(REAL8 input);
void swap_bytes(void *buf,INT4m buflen);
void swap_words4(void *buf,INT4m count);
void swap_words8(void *buf,INT4m count);

// inline codes...
inline UINT2 flip(UINT2 input)
{
  UCHAR temp;
  union { UCHAR byte[2]; UINT2 i; } word;
  word.i=input;
  temp=word.byte[0]; word.byte[0]=word.byte[1]; word.byte[1]=temp;
  return word.i;
}

inline INT2 flip(INT2 input)
{
  union { UINT2 u; INT2 i; } word;
  word.i=input;
  word.u=flip(word.u);  // signed <-> unsigned casting is
  return word.i;        // not well specified in C/C++
}

inline UINT4 flip(UINT4 input)
{
  UCHAR temp;
  union { UCHAR byte[4]; UINT4 i; } word;
  word.i=input;
  temp=word.byte[1]; word.byte[1]=word.byte[2]; word.byte[2]=temp;
  temp=word.byte[0]; word.byte[0]=word.byte[3]; word.byte[3]=temp;
  return word.i;
}

inline INT4 flip(INT4 input)
{
  union { UINT4 u; INT4 i; } word;
  word.i=input;  
  word.u=flip(word.u);  // signed <-> unsigned casting is
  return word.i;        // not well specified in C/C++
}

inline REAL4 flip(REAL4 input)
{
  union { UINT4 u; REAL4 r; } word;
  word.r=input;  
  word.u=flip(word.u);
  return word.r;
}

inline REAL8 flip(REAL8 input)
{ // This can be used on any 8 byte wide type
  UCHAR temp;
  union { UCHAR byte[8]; REAL8 r; } word;
  word.r=input;
  temp=word.byte[3]; word.byte[3]=word.byte[4]; word.byte[4]=temp;
  temp=word.byte[2]; word.byte[2]=word.byte[5]; word.byte[5]=temp;
  temp=word.byte[1]; word.byte[1]=word.byte[6]; word.byte[6]=temp;
  temp=word.byte[0]; word.byte[0]=word.byte[7]; word.byte[7]=temp;
  return word.r;
}

//////////////////////////////////////////////////////////////////////////
// String class
class String
{
  friend INT4m StringCompare(const String &,
			     const String &); // Replicates strcmp
  friend INT4m StringSpan(const String &,
			  const char *); // Replicates strspn
private:
  static const ClassDoc class_doc;
  char *str;
public:
  String(void);                // Default constructor
  String(const String &copyobj);     // Copy constructor
  String(const char *cptr);
  String& operator=(const String &rhs);  // Assignment operator
  String& operator=(const char *cptr);   // Ditto
  ~String();

  char& operator[](int offset) { return str[offset]; }
  // The following [] operator can be used on 'const String' objects.
  const char& operator[](int offset) const { return str[offset]; }

  operator const char*() const { return (const char *)str; }
  // Note: In the last 3 functions, the returned pointer may become
  //  invalid if later any of the non-const String member functions
  //  are called.

  // Length *not* including trailing null character.
  int Length(void) const { return strlen(str); }

  // Append functions; grows str as needed.  The char version appends
  // up to the first '\0' character or maxlen, which ever happens first.
  // Either way the last char in this->str is set to '\0'.
  // Both functions return *this.
  String& Append(const String& appendstr);
  String& Append(const char *appendarr,int maxlen);

  // Function to trim leading & trailing whitespace (as determined by
  // 'isspace' function) from string.
  String& Trim();
};

#endif // MISC_H

