/* 
 * oommfInit.c --
 *
 */

#include "machine.h"
#include "misc.h"
#include "oommf.h"

/*
 * The following string is the Tcl script executed in new 
 * Tcl 7.4 interpreters.  It defines some commands missing in
 * Tcl 7.4, but useful for interpreting portable Tcl scripts.
 */

//#if !((TCL_MAJOR_VERSION == 7) && (TCL_MINOR_VERSION >= 5))
static char filePathtypeScript[] =
"proc _OMF_FilePathtype {args} {\n\
   if {[llength args] != 1} {\n\
     return -code error \"wrong # args:\\\n\
should be \\\"file pathtype name\\\"\"\n\
   }\n\
   set path [lindex $args 0]\n\
   if [regexp {^[/~]} $path] {\n\
     return absolute\n\
   } else {\n\
     return relative\n\
   }\n\
}";

static char fileJoinScript[] =
"proc _OMF_FileJoin {args} {\n\
   if {[llength args] == 0} {\n\
     return -code error \"wrong # args:\\\n\
should be \\\"file join args\\\"\"\n\
   }\n\
   for {set i [expr [llength $args] - 1]} {$i > 0} {incr i -1} {\n\
     set element [lindex $args $i]\n\
     if {![string compare [file pathtype $element] absolute]} {break}\n\
   }\n\
   return [join [lrange $args $i [llength $args]] /]\n\
}";

static char fileSplitScript[] =
"proc _OMF_FileSplit {args} {\n\
   if {[llength args] != 1} {\n\
     return -code error \"wrong # args:\\\n\
should be \\\"file split name\\\"\"\n\
   }\n\
   set path [lindex $args 0]\n\
   set rawelements [split $path /]\n\
   set result {}\n\
   if [llength $rawelements] {\n\
     if {![string compare [lindex $rawelements 0] {}]} {\n\
       lappend result /\n\
     } else {\n\
       lappend result [lindex $rawelements 0]\n\
     }\n\
     foreach element [lrange $rawelements 1 [llength $rawelements]] {\n\
       if {![string compare [file pathtype $element] absolute]} {\n\
         lappend result ./$element\n\
       } else {\n\
         lappend result $element\n\
       }\n\
     }\n\
   }\n\
   return $result\n\
}";

static char patchScript[] =
"proc _OMF_Patch {} {\n\
   rename _OMF_Patch {}\n\
# pre-7.5 Tcl doesn't have bgerror. Alias it to tkerror. \n\
#	What if local Tcl installation defined one? \n\
#	This will override it.  Is that bad? \n\
   proc bgerror {args} {eval tkerror $args}\n\
# pre-7.5 Tcl only lives in Unix, so add a few portable file handling \n\
# routines to the file command suitable for Unix \n\
   rename file _OMF_File\n\
   proc file {command args} {\n\
     switch $command {\n\
       pathtype {eval _OMF_FilePathtype $args}\n\
       join	{eval _OMF_FileJoin $args}\n\
       split	{eval _OMF_FileSplit $args}\n\
       default	{eval _OMF_File $command $args}\n\
     }\n\
   }\n\
 }\n\
if {[string compare $tcl_version 7.4] == 0} {_OMF_Patch}\n\
#bgerror Patched!";
//#endif // !((TCL_MAJOR_VERSION == 7) && (TCL_MINOR_VERSION >= 5))

static char preBootScript[] =
"proc _OMF_PreBoot {} {\n\
   rename _OMF_PreBoot {}\n\
   global omf_library omf_debug\n\
   if {![info exists omf_debug]} {set omf_debug 0}\n\
   proc OMF_PackagePrefix {} {return {}}\n\
   proc OMFOOMMF_PackagePrefix {} {OMF_PackagePrefix}\n\
   proc OMF_PackageName {} {return Oommf}\n\
   global _omf_g\n\
   proc OMF_Version {} [list return $_omf_g(version)]\n\
   unset _omf_g(version)\n\
   proc OMF_PatchLevel {} [list return $_omf_g(patchLevel)]\n\
   unset _omf_g(patchLevel)\n\
   proc OMF_FileNameBase {} {string tolower [OMF_PackageName]}\n\
   proc OMF_SimplifyFilePath {path} {\n\
     while {[regsub {(/[.]/)|(/[.]$)} $path / path]} {}\n\
     while {[regsub [join { \n\
			(/[^/]/[.][.]/)|		\n\
			(/[^/][^/][^/]+/[.][.]/)|	\n\
			(/[^/][^/.]/[.][.]/)|		\n\
			(/[^/.][^/]/[.][.]/)|		\n\
			(/[^/]/[.][.]$)|		\n\
			(/[^/][^/][^/]+/[.][.]$)|	\n\
			(/[^/][^/.]/[.][.]$)|		\n\
			(/[^/.][^/]/[.][.]$)		\n\
		       } {}] $path / path]} {}		\n\
     regsub {/$} $path {} path\n\
     return $path\n\
   }\n\
}";

/*
 * The following string is the startup script executed in new
 * interpreters.  It looks on disk in several different directories
 * for a script "oommf.tcl" that is compatible with this version
 * of OOMMF.  The oommf.tcl script does all of the real work of
 * initialization.
 */

static char bootScript[] =
"proc _OMF_Boot {} {\n\
   rename _OMF_Boot {}\n\
   _OMF_PreBoot\n\
   if [catch {info nameofexecutable} execname] {\n\
     proc OMF_NameOfExecutable {} {\n\
       global argv0 env\n\
       set path [pwd]\n\
       if {[string compare [file pathtype [file dirname $argv0]] relative]} { \n\
         lappend path [file dirname $argv0]\n\
       } else {\n\
         lappend path [file dirname [file join [pwd] $argv0]]\n\
       }\n\
       if [info exists env(PATH)] {\n\
         set path [concat $path [split $env(PATH) :]]\n\
       }\n\
       foreach dir $path {\n\
         if [file executable [file join $dir [file tail $argv0]]] {\n\
           return [file join $dir [file tail $argv0]]\n\
         }\n\
       }\n\
       bgerror \"Can't determine name of executable!\"\n\
     }\n\
     set execname [OMF_NameOfExecutable]\n\
   }\n\
   set execname [OMF_SimplifyFilePath $execname]\n\
   if [info exists omf_library] {set library $omf_library}\n\
   set execdir [file dirname $execname]\n\
   catch {uplevel #0 source [file join $execdir [file rootname [file tail \\\n\
                             $execname]].tcl]}\n\
   if [info exists library] {\n\
     set omf_library $library\n\
   }\n\
}\n\
_OMF_Boot\n\
#bgerror Booted!";

/*
 * The following string is a script whic defines some procs used
 * byt the startup script below.  It's only broken out as a separate
 * script because come compilers don't like long string literals.
 */

static char preInitScript[] =
"proc OMF_AppendUniqueExistingReadableDirectory {listref dir} {\n\
   upvar $listref list\n\
   if {[file isdirectory $dir] && [file readable $dir]} {\n\
     if {[lsearch -exact $list $dir] < 0} {\n\
       lappend list $dir\n\
     }\n\
   }\n\
 }\n\
 proc OMF_ListLibraryCandidates {pfx PFX} {\n\
   global omf_library env tk_library\n\
   set dirs {}\n\
   set libvar [string toupper [OMF${PFX}_PackageName]]_LIBRARY\n\
   if [info exists env($libvar)] {\n\
     OMF_AppendUniqueExistingReadableDirectory dirs $env($libvar)\n\
   }\n\
   if [info exists omf${pfx}_library] {\n\
     OMF_AppendUniqueExistingReadableDirectory dirs [set omf${pfx}_library]\n\
   }\n\
   if [catch {info nameofexecutable} execname] {\n\
     set execname [OMF_NameOfExecutable]\n\
   }\n\
   set execname [OMF_SimplifyFilePath $execname]\n\
   set parentDir [file dirname [file dirname $execname]]\n\
   OMF_AppendUniqueExistingReadableDirectory dirs [file join $parentDir lib]\n\
   if [info exists omf_library] {\n\
     OMF_AppendUniqueExistingReadableDirectory dirs $omf_library\n\
   }\n\
# Restore this when version/patchlevel is in directory names\n\
#   if [string match {*[ab]*} [OMF${PFX}_PatchLevel]] {\n\
#     set lib [OMF${PFX}_FileNameBase][OMF${PFX}_PatchLevel]\n\
#   } else {\n\
#     set lib [OMF${PFX}_FileNameBase][OMF${PFX}_Version]\n\
#   }\n\
# Use this instead for now\n\
   set lib [OMF${PFX}_FileNameBase]\n\
#\n\
   OMF_AppendUniqueExistingReadableDirectory dirs \\\n\
     [file join [file dirname $parentDir] $lib library]\n\
   if [info exists omf_library] {\n\
     OMF_AppendUniqueExistingReadableDirectory dirs [file join \\\n\
       [file dirname [file dirname $omf_library]] $lib library]\n\
   }\n\
   OMF_AppendUniqueExistingReadableDirectory dirs [file dirname $tk_library]\n\
   OMF_AppendUniqueExistingReadableDirectory dirs \\\n\
     [file dirname [info library]]\n\
   return $dirs\n\
 }\n\
#bgerror Prepped!";

/*
 * The following string is the startup script executed in new
 * interpreters.  It looks on disk in several different directories
 * for a script "oommf.tcl" that is compatible with this version
 * of OOMMF.  The oommf.tcl script does all of the real work of
 * initialization.
 */

static char initScript[] =
"proc _OMF_Init {} {\n\
   rename _OMF_Init {}\n\
   proc OMF_SourceInitScript {pfx} {\n\
     global omf${pfx}_library\n\
     set PFX [string toupper $pfx]\n\
     set dirs [OMF_ListLibraryCandidates $pfx $PFX]\n\
     foreach i $dirs {\n\
       set omf${pfx}_library $i\n\
#       bgerror \"Sourcing [file join [set omf${pfx}_library] \\\n\
#                   [OMF${PFX}_FileNameBase][OMF${PFX}_Version] \\\n\
#                   [OMF${PFX}_FileNameBase].tcl] ...\"\n\
       if ![catch {uplevel #0 source [file join [set omf${pfx}_library] \\\n\
                   [OMF${PFX}_FileNameBase][OMF${PFX}_Version] \\\n\
                   [OMF${PFX}_FileNameBase].tcl]}] {\n\
#         bgerror \"Sourced!\"\n\
         return\n\
       }\n\
     }\n\
     set libvar [string toupper [OMF${PFX}_PackageName]]_LIBRARY\n\
     set msg \"[OMF${PFX}_PackageName] initialization failure.\n\n\"\n\
     append msg \"Can't find a usable initialization script, \"\n\
     append msg \"[file join [OMF${PFX}_FileNameBase][OMF${PFX}_Version] \\\n\
                             [OMF${PFX}_FileNameBase].tcl], \"\n\
     append msg \"in the following directories:\n\n\"\n\
     append msg \"[join $dirs {\n}]\n\n\"\n\
     append msg \"[OMF${PFX}_PackageName] probably hasn't been \"\n\
     append msg \"installed properly.\n\n\"\n\
     append msg \"Either (re-)install [OMF${PFX}_PackageName], \"\n\
     append msg \"or set the variable \"\n\
     append msg \"$libvar to the name \"\n\
     append msg \"of a directory containing \"\n\
     append msg \"[file join [OMF${PFX}_FileNameBase][OMF${PFX}_Version] \\\n\
                             [OMF${PFX}_FileNameBase].tcl], \"\n\
     bgerror $msg\n\
     exit\n\
   }\n\
   OMF_SourceInitScript {}\n\
}\n\
_OMF_Init\n\
#bgerror Initialized!";

/*
 *----------------------------------------------------------------------
 *
 * Oommf_Init --
 *
 * Results:
 *	Returns a standard Tcl result.  Leaves an error message or result
 *	in interp->result.
 *
 * Side effects:
 *	Sets standard global variables for the Oommf package.
 *	Runs "oommf.tcl" script.
 *
 *----------------------------------------------------------------------
 */

int
Oommf_Init(Tcl_Interp *interp)
{
#if (DEBUGLVL<1) || (SYSTEM_TYPE == WINDOWS)
  // Disable TCL puts command
  CHAR buf[SCRATCH_BUF_SIZE];
  strncpy(buf,"rename puts _OMF_OriginalPuts ; proc puts {args} {}",
	  sizeof(buf));
  if(Tcl_GlobalEval(interp,buf) != TCL_OK) return TCL_ERROR;
#else
  // Keep puts active
  CHAR buf[SCRATCH_BUF_SIZE];
  strncpy(buf,"rename puts _OMF_OriginalPuts ; "
	  "proc puts {args} {eval _OMF_OriginalPuts $args}",
	  sizeof(buf));
  if(Tcl_GlobalEval(interp,buf) != TCL_OK) return TCL_ERROR;
#endif // DEBUGLVL<1...
  // NOTE: _OMF_OriginalPuts should be used *ONLY* with an explicit
  //       filehandle, which must *not* be stdout or stderr.  The
  //       original puts proc is disabled because "puts stdout" and
  //       "puts stderr" will likely be unavailable under MS-Windows.

//#if !((TCL_MAJOR_VERSION == 7) && (TCL_MINOR_VERSION >= 5))
// Add some Tcl 7.5+ commands to pre-7.5 Tcl's so Oommf scripts
// can use more portable Tcl language (eg. 'file join').
    if (Tcl_Eval(interp, filePathtypeScript) != TCL_OK) {
      return TCL_ERROR;
    }
    if (Tcl_Eval(interp, fileJoinScript) != TCL_OK) {
      return TCL_ERROR;
    }
    if (Tcl_Eval(interp, fileSplitScript) != TCL_OK) {
      return TCL_ERROR;
    }
    if (Tcl_Eval(interp, patchScript) != TCL_OK) {
      return TCL_ERROR;
    }
//#endif

// Define the standard procs and global variables of the package
//	The first few from macros...
    Tcl_SetVar2(interp, "_omf_g", "patchLevel", 
                        OOMMF_PATCH_LEVEL, TCL_GLOBAL_ONLY);
    Tcl_SetVar2(interp, "_omf_g", "version", 
                        OOMMF_VERSION, TCL_GLOBAL_ONLY);
//	The rest from a boot script...
    if (Tcl_Eval(interp, preBootScript) != TCL_OK) {
      return TCL_ERROR;
    }
    if (Tcl_Eval(interp, bootScript) != TCL_OK) {
      return TCL_ERROR;
    }

// Make some preparations for the initialization script
    if (Tcl_Eval(interp, preInitScript) != TCL_OK) {
      return TCL_ERROR;
    }
// Finally, evaluate an initialization script
    return Tcl_Eval(interp, initScript);
}
